from typing import Union

from convertdate import hebrew  # type: ignore
from lark.exceptions import UnexpectedCharacters

from undate import Undate, UndateInterval
from undate.converters.base import BaseCalendarConverter
from undate.converters.calendars.hebrew.parser import hebrew_parser
from undate.converters.calendars.hebrew.transformer import HebrewDateTransformer


class HebrewDateConverter(BaseCalendarConverter):
    """
    Converter for Hebrew Anno Mundicalendar.

    Support for parsing Anno Mundi dates and converting to Undate and UndateInterval
    objects in the Gregorian calendar.
    """

    #: converter name: Hebrew
    name: str = "Hebrew"
    calendar_name: str = "Anno Mundi"

    #: earliest possible year in the Hebrew calendar is year 1, it does not go negative
    MIN_YEAR: int = 1
    # convertdate gives a month 34 for numpy max year 2.5^16, so scale it back a bit
    MAX_YEAR = int(2.5e12)

    #: arbitrary known non-leap year; 4816 is a non-leap year with 353 days (minimum possible)
    NON_LEAP_YEAR: int = 4816
    #: arbitrary known leap year; 4837 is a leap year with 385 days (maximum possible)
    LEAP_YEAR: int = 4837

    def __init__(self):
        self.transformer = HebrewDateTransformer()

    def min_month(self) -> int:
        """Smallest numeric month for this calendar."""
        return 1

    def max_month(self, year: int) -> int:
        """Maximum numeric month for this calendar. In Hebrew calendar, this is 12 or 13
        depending on whether it is a leap year."""
        return hebrew.year_months(year)

    def first_month(self) -> int:
        """First month in this calendar. The Hebrew civil year starts in Tishri."""
        return hebrew.TISHRI

    def last_month(self, year: int) -> int:
        """Last month in this calendar. Hebrew civil year starts in Tishri,
        Elul is the month before Tishri."""
        return hebrew.ELUL

    def max_day(self, year: int, month: int) -> int:
        """maximum numeric day for the specified year and month in this calendar"""
        # NOTE: unreleased v2.4.1 of convertdate standardizes month_days to month_length
        return hebrew.month_days(year, month)

    def days_in_year(self, year: int) -> int:
        """the number of days in the specified year for this calendar"""
        return int(hebrew.year_days(year))

    def representative_years(self, years: None | list[int] = None) -> list[int]:
        """Takes a list of years and returns a subset with all possible variations in number of days.
        If no years are specified, returns ...
        """

        year_lengths = set()
        max_year_lengths = 6  # there are 6 different possible length years

        # if years is unset or list is empty
        if not years:
            # NOTE: this does not cover all possible lengths, but should cover min/max
            return [self.LEAP_YEAR, self.NON_LEAP_YEAR]

        rep_years = []
        for year in years:
            days = self.days_in_year(year)
            if days not in year_lengths:
                year_lengths.add(days)
                rep_years.append(year)

            # stop if we find one example of each type of year
            if len(year_lengths) == max_year_lengths:
                break

        return rep_years

    def to_gregorian(self, year: int, month: int, day: int) -> tuple[int, int, int]:
        """Convert a Hebrew date, specified by year, month, and day,
        to the Gregorian equivalent date. Returns a tuple of year, month, day.
        """
        return hebrew.to_gregorian(year, month, day)

    def parse(self, value: str) -> Union[Undate, UndateInterval]:
        """
        Parse a Hebrew date string and return an :class:`~undate.undate.Undate` or
        :class:`~undate.undate.UndateInterval`.
        The Hebrew date string is preserved in the undate label.
        """
        if not value:
            raise ValueError("Parsing empty string is not supported")

        # parse the input string, then transform to undate object
        try:
            # parse the string with our Hebrew date parser
            parsetree = hebrew_parser.parse(value)
            # transform the parse tree into an undate or undate interval
            undate_obj = self.transformer.transform(parsetree)
            # set the original date as a label, with the calendar name
            undate_obj.label = f"{value} {self.calendar_name}"
            return undate_obj
        except UnexpectedCharacters as err:
            raise ValueError(f"Could not parse '{value}' as a Hebrew date") from err

    # do we need to support conversion the other direction?
    # i.e., generate a Hebrew date from an abitrary undate or undate interval?



from typing import Optional, Union

from lark.exceptions import UnexpectedCharacters

from undate import Undate, UndateInterval
from undate.converters.base import BaseDateConverter
from undate.converters.edtf.parser import edtf_parser
from undate.converters.edtf.transformer import EDTFTransformer
from undate.date import DatePrecision


#: character for unspecified digits
EDTF_UNSPECIFIED_DIGIT: str = "X"


class EDTFDateConverter(BaseDateConverter):
    """
    Converter for Extended Date/Time Format (EDTF).

    Supports parsing and serializing dates and date ranges in EDTF format.
    Does not support all of EDTF, and only supports dates and not times.
    """

    #: converter name: EDTF
    name: str = "EDTF"

    def __init__(self):
        self.transformer = EDTFTransformer()

    def parse(self, value: str) -> Union[Undate, UndateInterval]:
        """
        Parse a string in a supported EDTF date or date interval format and
        return an :class:`~undate.undate.Undate` or
        :class:`~undate.undate.UndateInterval`.
        """
        if not value:
            raise ValueError("Parsing empty/unset string is not supported")

        # parse the input string, then transform to undate object
        try:
            parsetree = edtf_parser.parse(value)
            return self.transformer.transform(parsetree)
        except UnexpectedCharacters:
            raise ValueError(
                "Parsing failed: '%s' is not a supported EDTF date format" % value
            )

    def _convert_missing_digits(
        self, value: Optional[str], old_missing_digit: str
    ) -> Optional[str]:
        if value:
            return value.replace(old_missing_digit, EDTF_UNSPECIFIED_DIGIT)
        return None

    def to_string(self, undate: Union[Undate, UndateInterval]) -> str:
        """
        Convert an :class:`~undate.undate.Undate` or
        :class:`~undate.undate.UndateInterval` to EDTF format.
        """
        if isinstance(undate, Undate):
            return self._undate_to_string(undate)
        elif isinstance(undate, UndateInterval):
            # NOTE: what is the difference between an open interval and unknown start/end?
            # spec distinguishes between these, open is ".." but unknown is ""
            start = self._undate_to_string(undate.earliest) if undate.earliest else ".."
            end = self._undate_to_string(undate.latest) if undate.latest else ".."
            return f"{start}/{end}"

    def _undate_to_string(self, undate: Undate) -> str:
        # in theory it's possible to use the parser and reconstruct using a tree,
        # but that seems much more complicated and would be harder to read
        parts = []

        if undate.precision >= DatePrecision.YEAR:
            year = self._convert_missing_digits(undate.year, undate.MISSING_DIGIT)
            # years with more than 4 digits should be prefixed with Y
            # (don't count minus sign when checking digits)
            if year and len(year.lstrip("-")) > 4:
                year = f"Y{year}"

            # TODO: handle uncertain / approximate
            parts.append(year or EDTF_UNSPECIFIED_DIGIT * 4)

        # beware when we add more date precisions,
        # week-level won't necessarily mean we know the month

        if undate.precision >= DatePrecision.MONTH:
            # TODO: handle uncertain / approximate
            parts.append(
                self._convert_missing_digits(undate.month, undate.MISSING_DIGIT)
                or EDTF_UNSPECIFIED_DIGIT * 2
            )

        if undate.precision >= DatePrecision.DAY:
            # TODO: handle uncertain / approximate
            parts.append(
                self._convert_missing_digits(undate.day, undate.MISSING_DIGIT)
                or EDTF_UNSPECIFIED_DIGIT * 2
            )

        if parts:
            return "-".join(parts)

        # how can we have an empty string? probably shouldn't get here
        raise ValueError("Failed to generate an EDTF string from %r", undate)



import pytest

from undate import Undate, UndateInterval
from undate.converters.edtf.parser import edtf_parser
from undate.converters.edtf.transformer import EDTFTransformer

# for now, just test that valid dates can be parsed

testcases = [
    ("1984", Undate(1984)),
    ("1984-05", Undate(1984, 5)),
    ("1984-12", Undate(1984, 12)),
    ("1001-03-30", Undate(1001, 3, 30)),
    ("1000/2000", UndateInterval(Undate(1000), Undate(2000))),
    ("1000-01/2000-05-01", UndateInterval(Undate(1000, 1), Undate(2000, 5, 1))),
    # level 1
    ("Y17000002", Undate(17000002)),
    ("Y-17000002", Undate(-17000002)),
    # negative year
    ("-1985", Undate(-1985)),
    # "2001-21",  # spring 2001
    # qualifiers TODO - not yet supported by undate
    # "1984?",
    # "2004-06~",
    # "2004-06-11%",
    # # unspecified digits from right
    ("201X", Undate("201X")),
    ("20XX", Undate("20XX")),
    ("2004-XX", Undate(2004, "XX")),
    ("1985-04-XX", Undate(1985, 4, "XX")),
    ("1985-XX-XX", Undate(1985, "XX", "XX")),
    # # open ended intervals
    ("1985-04-12/..", UndateInterval(Undate(1985, 4, 12), None)),
    ("1985-04/..", UndateInterval(Undate(1985, 4), None)),
    ("../1985-04-12", UndateInterval(None, Undate(1985, 4, 12))),
    ("/1985-04-12", UndateInterval(None, Undate(1985, 4, 12))),
    # level 2 unspecified digits
    ("156X-12-25", Undate("156X", 12, 25)),
    ("XXXX-12-XX", Undate("XXXX", 12, "XX")),
    ("1XXX-XX", Undate("1XXX", "XX")),
    ("1984-1X", Undate(1984, "1X")),
]


@pytest.mark.parametrize("date_string,expected", testcases)
def test_transform(date_string, expected):
    transformer = EDTFTransformer()
    # parse the input string, then transform to undate object
    parsetree = edtf_parser.parse(date_string)
    # since the same unknown date is not considered strictly equal,
    # compare object representations
    assert repr(transformer.transform(parsetree)) == repr(expected)



import pathlib

from lark import Lark

grammar_path = pathlib.Path(__file__).parent / "hebrew.lark"

with open(grammar_path) as grammar:
    # NOTE: LALR parser is faster but can't be used to ambiguity between years and dates
    hebrew_parser = Lark(grammar.read(), start="hebrew_date", strict=True)



from calendar import monthrange, isleap

from undate.converters.base import BaseCalendarConverter


class GregorianDateConverter(BaseCalendarConverter):
    """
    Calendar converter class for Gregorian calendar.
    """

    #: converter name: Gregorian
    name: str = "Gregorian"
    #: calendar
    calendar_name: str = "Gregorian"

    #: arbitrary known non-leap year
    NON_LEAP_YEAR: int = 2022
    #: arbitrary known leap year
    LEAP_YEAR: int = 2024

    def min_month(self) -> int:
        """First month for the Gregorian calendar."""
        return 1

    def max_month(self, year: int) -> int:
        """maximum numeric month for the specified year in the Gregorian calendar"""
        return 12

    def max_day(self, year: int, month: int) -> int:
        """maximum numeric day for the specified year and month in this calendar"""
        # if month is known, use that to calculate
        if month:
            # if year is known, use it; otherwise use a known non-leap year
            # (only matters for February)
            year = year or self.NON_LEAP_YEAR

            # Use monthrange from python builtin calendar module.
            # returns first day of the month and number of days in the month
            # for the specified year and month.
            _, max_day = monthrange(year, month)
        else:
            # if year and month are unknown, return maximum possible
            # TODO: should this return an IntervalRange?
            max_day = 31

        return max_day

    def representative_years(self, years: None | list[int] = None) -> list[int]:
        """Takes a list of years and returns a subset with one leap year and one non-leap year.
        If no years are specified, returns a known leap year and non-leap year.
        """

        # if years is unset or list is empty
        if not years:
            return [self.LEAP_YEAR, self.NON_LEAP_YEAR]

        found_leap = False
        found_non_leap = False
        rep_years = []
        for year in years:
            if isleap(year):
                if not found_leap:
                    found_leap = True
                    rep_years.append(year)
            else:
                if not found_non_leap:
                    found_non_leap = True
                    rep_years.append(year)
            # stop as soon as we've found one example of each type of year
            if found_leap and found_non_leap:
                break

        return rep_years

    def to_gregorian(self, year, month, day) -> tuple[int, int, int]:
        """Convert to Gregorian date. This returns the specified by year, month,
        and day unchanged, but is provided for consistency since all calendar
        converters need to support conversion to Gregorian calendar for
        a common point of comparison.
        """
        return (year, month, day)



__version__ = "0.5.2"

from undate.date import DatePrecision, UnDelta
from undate.undate import Undate, Calendar
from undate.interval import UndateInterval

__all__ = [
    "Undate",
    "UndateInterval",
    "Calendar",
    "DatePrecision",
    "UnDelta",
    "__version__",
]



from undate.converters.calendars.gregorian import GregorianDateConverter
from undate.converters.calendars.hebrew import HebrewDateConverter
from undate.converters.calendars.islamic import IslamicDateConverter
from undate.converters.calendars.seleucid import SeleucidDateConverter

__all__ = [
    "GregorianDateConverter",
    "HebrewDateConverter",
    "IslamicDateConverter",    
    "SeleucidDateConverter",
]



"""
:class:`~undate.converters.BaseDateConverter` provides a base class for
implementing date converters, which can provide support for
parsing and generating dates in different formats.
The converter subclass :class:`undate.converters.BaseCalendarConverter`
provides additional functionality needed for calendar conversion.

To add support for a new date converter:

- Create a new file under ``undate/converters/``
    - For converters with sufficient complexity, you may want to create a submodule;
      see ``undate.converters.edtf`` for an example.
- Extend ``BaseDateConverter`` and implement ``parse`` and ``to_string`` methods
  as desired/appropriate for your converter
- Add unit tests for the new converter in ``tests/test_converters/``
- Optionally, you may want to create a notebook to demonstrate the use and value
  of the new converter.

The new subclass should be loaded automatically and included in the converters
returned by :meth:`BaseDateConverter.available_converters`

To add support for a new calendar converter:

- Create a new file under ``undate/converters/calendars/``
    - For converters with sufficient complexity, you may want to create a submodule;
      see ``undate.converters.calendars.islamic`` for an example.
- Extend ``BaseCalendarConverter`` and implement ``parse`` and ``to_string``
  formatter methods as desired/appropriate for your converter as well as the
  additional methods for ``max_month``, ``max_day``, and conversion ``to_gregorian``
  calendar.
- Import your calendar in ``undate/converters/calendars/__init__.py`` and include in `__all__``
- Add unit tests for the new calendar logic under ``tests/test_converters/calendars/``
- Add the new calendar to the ``Calendar`` enum of supported calendars in
  ``undate/undate.py`` and confirm that the `get_converter` method loads your
  calendar converter correctly (an existing unit test should cover this).
- Consider creating a notebook to demonstrate the use of the calendar
  converter.

Calendar converter subclasses are also automatically loaded and included
in the list of available converters.

-------------------
"""

import importlib
import logging
import pkgutil
from functools import cache
from typing import Dict, Type

from undate.date import Date

logger = logging.getLogger(__name__)


class BaseDateConverter:
    """Base class for parsing, formatting, and converting dates to handle
    specific formats and different calendars."""

    #: Converter name. Subclasses must define a unique name.
    name: str = "Base Converter"

    # provisional...
    LEAP_YEAR = 0
    NON_LEAP_YEAR = 0

    def parse(self, value: str):
        """
        Parse a string and return an :class:`~undate.undate.Undate` or
        :class:`~undate.undate.UndateInterval`. Must be implemented by
        subclasses.
        """
        # can't add type hint here because of circular import
        # should return an undate or undate interval
        raise NotImplementedError

    def to_string(self, undate) -> str:
        """
        Convert an :class:`~undate.undate.Undate` or
        :class:`~undate.undate.UndateInterval` to string.
        Must be implemented by subclasses.
        """

        # undate param should be of type Union[Undate, UndateInterval] but can't add type hint here because of circular import
        # convert an undate or interval to string representation for this format
        raise NotImplementedError

    # cache import class method to ensure we only import once
    @classmethod
    @cache
    def import_converters(cls) -> int:
        """Import all undate converters
        so that they will be included in available converters
        even if not explicitly imported. Only import once.
        returns the count of modules imported."""

        logger.debug("Loading converters under undate.converters")
        import undate.converters

        # load packages under this path with curent package prefix
        converter_path = undate.converters.__path__
        converter_prefix = f"{undate.converters.__name__}."

        import_count = 0
        for importer, modname, ispkg in pkgutil.iter_modules(
            converter_path, converter_prefix
        ):
            # import everything except the current file
            if not modname.endswith(".base"):
                importlib.import_module(modname)
                import_count += 1

        return import_count

    @classmethod
    def available_converters(cls) -> Dict[str, Type["BaseDateConverter"]]:
        """
        Dictionary of available converters keyed on name.
        """
        return {c.name: c for c in cls.subclasses()}  # type: ignore

    @classmethod
    def subclasses(cls) -> set[Type["BaseDateConverter"]]:
        """
        Set of available converters classes. Includes descendant
        subclasses, including calendar converters, but does not include
        :class:`BaseCalendarConverter`.
        """
        # ensure undate converters are imported
        cls.import_converters()

        # find all direct subclasses, excluding base calendar converter
        direct_subclasses = cls.__subclasses__()
        all_subclasses = set(direct_subclasses)
        # recurse to find nested subclasses
        for subc in direct_subclasses:
            all_subclasses |= subc.subclasses()

        # omit the calendar converter base class, which is not itself a converter
        all_subclasses -= {BaseCalendarConverter}
        return all_subclasses


class BaseCalendarConverter(BaseDateConverter):
    """Base class for calendar converters, with additional methods required
    for calendars."""

    #: Converter name. Subclasses must define a unique name.
    name: str = "Base Calendar Converter"

    #: arbitrary known non-leap year
    NON_LEAP_YEAR: int
    #: arbitrary known leap year
    LEAP_YEAR: int

    # minimum year for this calendar, if there is one
    MIN_YEAR: None | int = None
    # maximum year for this calendar, if there is one
    MAX_YEAR: None | int = None

    def min_month(self) -> int:
        """Smallest numeric month for this calendar."""
        raise NotImplementedError

    def max_month(self, year: int) -> int:
        """Maximum numeric month for this calendar"""
        raise NotImplementedError

    def first_month(self) -> int:
        """first month in this calendar; by default, returns :meth:`min_month`."""
        return self.min_month()

    def last_month(self, year: int) -> int:
        """last month in this calendar; by default, returns :meth:`max_month`."""
        return self.max_month(year)

    def max_day(self, year: int, month: int) -> int:
        """maximum numeric day for the specified year and month in this calendar"""
        raise NotImplementedError

    def days_in_year(self, year: int) -> int:
        """Number of days in the specified year in this calendar. The default implementation
        uses min and max month and max day methods along with Gregorian conversion method
        to calculate the number of days in the specified year.
        """
        year_start = Date(*self.to_gregorian(year, self.min_month(), 1))
        last_month = self.max_month(year)
        year_end = Date(
            *self.to_gregorian(year, last_month, self.max_day(year, last_month))
        )
        # add 1 because the difference doesn't include the end point
        return (year_end - year_start).days + 1

    def representative_years(self, years: None | list[int] = None) -> list[int]:
        """Returns a list of representative years within the specified list.
        Result should include one for each type of variant year for this
        calendar (e.g., leap year and non-leap year). If no years are specified,
        returns a list of representative years for the current calendar.
        """
        raise NotImplementedError

    def to_gregorian(self, year, month, day) -> tuple[int, int, int]:
        """Convert a date for this calendar specified by numeric year, month, and day,
        into the Gregorian equivalent date. Should return a tuple of year, month, day.
        """
        raise NotImplementedError



from typing import Union

from convertdate import islamic  # type: ignore
from lark.exceptions import UnexpectedCharacters

from undate import Undate, UndateInterval
from undate.converters.base import BaseCalendarConverter
from undate.converters.calendars.islamic.parser import islamic_parser
from undate.converters.calendars.islamic.transformer import IslamicDateTransformer


class IslamicDateConverter(BaseCalendarConverter):
    """
    Converter for Islamic Hijri calendar.

    Support for parsing Islamic Hijri dates and converting to Undate and UndateInterval
    objects in the Gregorian calendar.
    """

    #: converter name: Islamic
    name: str = "Islamic"
    calendar_name: str = "Islamic"

    #: arbitrary known non-leap year
    NON_LEAP_YEAR: int = 1457
    #: arbitrary known leap year
    LEAP_YEAR: int = 1458

    # minimum year for islamic calendar is 1 AH, does not go negative
    MIN_YEAR: None | int = 1
    # convertdate gives a month 34 for numpy max year 2.5^16, so scale it back a bit
    MAX_YEAR = int(2.5e12)

    def __init__(self):
        self.transformer = IslamicDateTransformer()

    def max_day(self, year: int, month: int) -> int:
        """maximum numeric day for the specified year and month in this calendar"""
        return islamic.month_length(year, month)

    def min_month(self) -> int:
        """smallest numeric month for this calendar."""
        return 1

    def max_month(self, year: int) -> int:
        """maximum numeric month for this calendar"""
        return 12

    def representative_years(self, years: None | list[int] = None) -> list[int]:
        """Takes a list of years and returns a subset with one leap year and one non-leap year.
        If no years are specified, returns a known leap year and non-leap year.
        """

        # if years is unset or list is empty
        if not years:
            return [self.LEAP_YEAR, self.NON_LEAP_YEAR]
        found_leap = False
        found_non_leap = False
        rep_years = []
        for year in years:
            if islamic.leap(year):
                if not found_leap:
                    found_leap = True
                    rep_years.append(year)
            else:
                if not found_non_leap:
                    found_non_leap = True
                    rep_years.append(year)
            # stop as soon as we've found one example of each type of year
            if found_leap and found_non_leap:
                break

        return rep_years

    def to_gregorian(self, year: int, month: int, day: int) -> tuple[int, int, int]:
        """Convert a Hijri date, specified by year, month, and day,
        to the Gregorian equivalent date. Returns a tuple of year, month, day.
        """
        # NOTE: this results in weird numbers for months when year gets sufficiently high
        return islamic.to_gregorian(year, month, day)

    def parse(self, value: str) -> Union[Undate, UndateInterval]:
        """
        Parse an Islamic/Hijri date string and return an :class:`~undate.undate.Undate` or
        :class:`~undate.undate.UndateInterval`.
        The Islamic/Hijri date string is preserved in the undate label.
        """
        if not value:
            raise ValueError("Parsing empty string is not supported")

        # parse the input string, then transform to undate object
        try:
            # parse the string with our Islamic Hijri date parser
            parsetree = islamic_parser.parse(value)
            # transform the parse tree into an undate or undate interval
            undate_obj = self.transformer.transform(parsetree)
            # set the original date as a label, with the calendar name
            undate_obj.label = f"{value} {self.calendar_name}"
            return undate_obj
        except UnexpectedCharacters as err:
            raise ValueError(f"Could not parse '{value}' as an Islamic date") from err

    # do we need to support conversion the other direction?
    # i.e., generate an Islamic Hijri date from an arbitrary undate or undate interval?



from undate.converters.calendars import GregorianDateConverter


class TestGregorianDateConverter:
    def test_to_gregorian(self):
        converter = GregorianDateConverter()
        # conversion is a no-op, returns values unchanged
        assert converter.to_gregorian(2025, 6, 15) == (2025, 6, 15)

    def test_min_month(self):
        assert GregorianDateConverter().min_month() == 1

    def test_max_month(self):
        assert GregorianDateConverter().max_month(2025) == 12

    def test_max_day(self):
        converter = GregorianDateConverter()
        assert converter.max_day(2025, 1) == 31
        assert converter.max_day(2025, 2) == 28
        assert converter.max_day(converter.LEAP_YEAR, 2) == 29
        assert converter.max_day(2025, 12) == 31

    def test_representative_years(self):
        converter = GregorianDateConverter()
        # single year is not filtered
        assert converter.representative_years([2025]) == [2025]
        # multiple non-leap years, returns just the first
        assert converter.representative_years([2025, 2026]) == [2025]
        # next leap year is 2028; returns first leap year and first non-leap year, in input order
        assert converter.representative_years([2025, 2026, 2028, 2029]) == [2025, 2028]

        # if no years are provided, returns a known leap year and non-leap year
        assert converter.representative_years() == [
            converter.LEAP_YEAR,
            converter.NON_LEAP_YEAR,
        ]
        assert converter.representative_years([]) == [
            converter.LEAP_YEAR,
            converter.NON_LEAP_YEAR,
        ]



import pytest
from undate.converters.calendars.islamic.parser import islamic_parser
from undate.converters.calendars.islamic.transformer import (
    IslamicDateTransformer,
    IslamicUndate,
)
from undate.undate import Undate, Calendar
from undate.date import DatePrecision


def test_islamic_undate():
    assert IslamicUndate(848).calendar == Calendar.ISLAMIC


testcases = [
    # examples from Princeton Geniza Project
    # date conversions checked with https://www.muqawwim.com/
    # Monday, 7 Jumādā I 1243 Hijrī (26 November, 1827 CE); Jumada I = month 5
    ("7 Jumādā I 1243", IslamicUndate(1243, 5, 7), DatePrecision.DAY),
    ("Jumādā I 1243", IslamicUndate(1243, 5), DatePrecision.MONTH),
    ("1243", IslamicUndate(1243), DatePrecision.YEAR),
    # Gregorian:  UndateInterval(Undate(1827, 7, 25), Undate(1828, 7, 13)),
    # Zū al-Qaʿdah / Dhu l-Qa'da = month 11
    ("27 Dhū l-Qaʿda 632", IslamicUndate(632, 11, 27), DatePrecision.DAY),
    # Rajab = month 7
    ("Rajab 495", IslamicUndate(495, 7), DatePrecision.MONTH),
    ("441", IslamicUndate(441), DatePrecision.YEAR),
    # examples from ISMI data (reformatted to day month year)
    # Rabi 1 = month 3
    ("14 Rabīʿ I 901", IslamicUndate(901, 3, 14), DatePrecision.DAY),
    ("Rabīʿ I 490", IslamicUndate(490, 3), DatePrecision.MONTH),    
    ("884", IslamicUndate(884), DatePrecision.YEAR),
    # Gregorian:  UndateInterval(Undate(1479, 4, 3), Undate(1480, 3, 21)),
    # add when we support parsing ranges:
    # 900 Muḥarram 1 - 999 Ḏu al-Ḥijjaẗ 29 : 1494-10-11 to 1591-10-18
]


@pytest.mark.parametrize("date_string,expected,expected_precision", testcases)
def test_transform(date_string, expected, expected_precision):
    transformer = IslamicDateTransformer(visit_tokens=True)
    # parse the input string, then transform to undate object
    parsetree = islamic_parser.parse(date_string)
    transformed_date = transformer.transform(parsetree)
    assert transformed_date == expected
    # currently only undates have date precision
    if isinstance(transformed_date, Undate):
        assert transformed_date.precision == expected_precision
    # transformer doesn't have access to date string,
    # label will need to be set by the converter class



# Pre 3.10 requires Union for multiple types, e.g. Union[int, None] instead of int | None
from typing import Optional, Union


from undate import Undate
from undate.date import ONE_DAY, ONE_YEAR, Timedelta
from undate.converters.base import BaseDateConverter


class UndateInterval:
    """A date range between two uncertain dates.

    :param earliest: Earliest undate
    :type earliest: `undate.Undate`
    :param latest: Latest undate
    :type latest:  `undate.Undate`
    :param label: A string to label a specific undate interval, similar to labels of `undate.Undate`.
    :type label: `str`
    """

    # date range between two undates
    earliest: Union[Undate, None]
    latest: Union[Undate, None]
    label: Union[str, None]

    # TODO: think about adding an optional precision / length /size field
    # using DatePrecision for intervals of any standard duration (decade, century)

    def __init__(
        self,
        earliest: Optional[Undate] = None,
        latest: Optional[Undate] = None,
        label: Optional[str] = None,
    ):
        # takes two undate objects; allows conversion from supported types
        if earliest:
            try:
                earliest = Undate.to_undate(earliest)
            except TypeError as err:
                raise ValueError(
                    f"earliest date {earliest} cannot be converted to Undate"
                ) from err
        if latest:
            try:
                latest = Undate.to_undate(latest)
            except TypeError as err:
                raise ValueError(
                    f"latest date {latest} cannot be converted to Undate"
                ) from err

        # check that the interval is valid
        if latest and earliest and latest <= earliest:
            raise ValueError(f"invalid interval {earliest}-{latest}")

        self.earliest = earliest
        self.latest = latest
        self.label = label

    def __str__(self) -> str:
        # using EDTF syntax for open ranges
        return "%s/%s" % (self.earliest or "..", self.latest or "")

    def format(self, format) -> str:
        """format this undate interval as a string using the specified format;
        for now, only supports named converters"""
        converter_cls = BaseDateConverter.available_converters().get(format, None)
        if converter_cls:
            return converter_cls().to_string(self)

        raise ValueError(f"Unsupported format '{format}'")

    def __repr__(self) -> str:
        init_opts = {
            "earliest": repr(self.earliest) if self.earliest else None,
            "latest": repr(self.latest) if self.latest else None,
            "label": f"{self.label!r}" if self.label else None,
        }
        init_str = ", ".join(
            [f"{key}={val}" for key, val in init_opts.items() if val is not None]
        )
        return f"undate.UndateInterval({init_str})"

    def __eq__(self, other) -> bool:
        # currently doesn't support comparison with any other types
        if not isinstance(other, UndateInterval):
            return NotImplemented
        # consider interval equal if both dates are equal
        return self.earliest == other.earliest and self.latest == other.latest

    def duration(self) -> Timedelta:
        """Calculate the duration between two undates.
        Note that durations are inclusive (i.e., a closed interval), and
        include both the earliest and latest date rather than the difference
        between them.

        :returns: A duration
        :rtype: Timedelta
        """
        # what is the duration of this date range?

        # if range is open-ended, can't calculate
        if self.earliest is None or self.latest is None:
            raise NotImplementedError(
                "Cannot calculate duration for open-ended interval"
            )

        # if both years are known, subtract end of range from beginning of start
        if self.latest.known_year and self.earliest.known_year:
            return self.latest.latest - self.earliest.earliest + ONE_DAY

        # if neither year is known...
        elif not self.latest.known_year and not self.earliest.known_year:
            # under what circumstances can we assume that if both years
            # are unknown the dates are in the same year or sequential?
            duration = self.latest.earliest - self.earliest.earliest
            # if we get a negative, we've wrapped from end of one year
            # to the beginning of the next;
            # recalculate assuming second date is in the subsequent year
            if duration.days < 0:
                end = self.latest.earliest + ONE_YEAR
                duration = end - self.earliest.earliest

            # add the additional day *after* checking for a negative
            # or after recalculating with adjusted year
            duration += ONE_DAY

            return duration

        else:
            # is there any meaningful way to calculate duration
            # if one year is known and the other is not?
            raise NotImplementedError

    def __contains__(self, other: object) -> bool:
        """Determine if another interval or date falls within this
        interval.  Supports comparison with :class:`UndateInterval`
        or anything that can be converted with :meth:`Undate.to_undate`."""
        # support comparison with another interval or anything
        # that can be converted to an Undate
        if isinstance(other, UndateInterval):
            # compare based on earliest/latest bounds
            other_earliest = other.earliest
            other_latest = other.latest
        else:
            # otherwise, try to convert to an Undate
            try:
                other = Undate.to_undate(other)
                other_latest = other_earliest = other
            except TypeError:
                # if conversion fails, then we don't support comparison
                raise

        # if either bound of the current interval is None,
        # then it is an open interval and we don't need to check the other value.
        # if the other value is set, then check that it falls within the
        # bounds of this interval
        return (
            self.earliest is None
            or other_earliest is not None
            and other_earliest >= self.earliest
        ) and (
            self.latest is None
            or other_latest is not None
            and other_latest <= self.latest
        )

    def intersection(self, other: "UndateInterval") -> Optional["UndateInterval"]:
        """Determine the intersection or overlap between two :class:`UndateInterval`
        objects and return a new interval. Returns None if there is no overlap.
        """
        try:
            # when both values are defined, return the inner bounds;
            # if not, return whichever is not None, or None
            earliest = (
                max(self.earliest, other.earliest)
                if self.earliest and other.earliest
                else self.earliest or other.earliest
            )
            latest = (
                min(self.latest, other.latest)
                if self.latest and other.latest
                else self.latest or other.latest
            )

            # if this results in an invalid interval, initialization
            # will throw an exception
            return UndateInterval(earliest, latest)
        except ValueError:
            return None



from lark import Transformer, Tree

from undate import Undate, Calendar


class IslamicUndate(Undate):
    """Undate convience subclass; sets default calendar to Islamic."""

    calendar = Calendar.ISLAMIC


class IslamicDateTransformer(Transformer):
    """Transform an Islamic Hijri date parse tree and return an Undate or
    UndateInterval."""

    def islamic_date(self, items):
        parts = {}
        for child in items:
            if child.data in ["year", "month", "day"]:
                # in each case we expect one integer value;
                # anonymous tokens convert to their value and cast as int
                value = int(child.children[0])
                parts[str(child.data)] = value

        # initialize and return an undate with islamic year, month, day and
        # islamic calendar
        return IslamicUndate(**parts)

    # year translation is not needed since we want a tree with name year
    # this is equivalent to a no-op
    # def year(self, items):
    #     return Tree(data="year", children=[items[0]])

    def month(self, items):
        # month has a nested tree for the rule and the value
        # the name of the rule (month_1, month_2, etc) gives us the
        # number of the month needed for converting the date
        tree = items[0]
        month_n = tree.data.split("_")[-1]
        return Tree(data="month", children=[month_n])



import logging

import pytest
from undate.converters.base import BaseDateConverter, BaseCalendarConverter
from undate.converters.calendars import (
    GregorianDateConverter,
    HebrewDateConverter,
    IslamicDateConverter,
)


class TestBaseDateConverter:
    def test_available_converters(self):
        available_converters = BaseDateConverter.available_converters()
        assert isinstance(available_converters, dict)

        # NOTE: import _after_ generating available formatters
        # so we can confirm it gets loaded
        from undate.converters.iso8601 import ISO8601DateFormat

        assert ISO8601DateFormat.name in available_converters
        assert available_converters[ISO8601DateFormat.name] == ISO8601DateFormat

    def test_converters_are_unique(self):
        assert len(BaseDateConverter.available_converters()) == len(
            BaseDateConverter.subclasses()
        ), "Formatter names have to be unique."

    def test_parse_not_implemented(self):
        with pytest.raises(NotImplementedError):
            BaseDateConverter().parse("foo bar baz")

    def test_parse_to_string(self):
        with pytest.raises(NotImplementedError):
            BaseDateConverter().to_string(1991)

    def test_subclasses(self):
        # define a nested subclass
        class SubSubConverter(IslamicDateConverter):
            pass

        subclasses = BaseDateConverter.subclasses()
        assert BaseCalendarConverter not in subclasses
        assert IslamicDateConverter in subclasses
        assert HebrewDateConverter in subclasses
        assert GregorianDateConverter in subclasses
        assert SubSubConverter in subclasses


def test_import_converters_import_only_once(caplog):
    # clear the cache, since any instantiation of an Undate
    # object anywhere in the test suite will populate it
    BaseDateConverter.import_converters.cache_clear()

    # run first, and confirm it runs and loads formatters
    with caplog.at_level(logging.DEBUG):
        import_count = BaseDateConverter.import_converters()
    # should import at least one thing (iso8601)
    assert import_count >= 1
    # should have log entry
    assert "Loading converters" in caplog.text

    # if we clear the log and run again, should not do anything
    caplog.clear()
    with caplog.at_level(logging.DEBUG):
        BaseDateConverter.import_converters()
    assert "Loading converters" not in caplog.text


@pytest.mark.last
def test_converters_unique_error():
    # confirm that unique converter check fails when it should

    # run this test last because we can't undefine the subclass
    # once it exists...
    class ISO8601DateFormat2(BaseDateConverter):
        name = "ISO8601"  # duplicates existing formatter

    assert len(BaseDateConverter.available_converters()) != len(
        BaseDateConverter.subclasses()
    )


class TestBaseCalendarConverter:
    def test_not_implemented(self):
        with pytest.raises(NotImplementedError):
            BaseCalendarConverter().min_month()
        with pytest.raises(NotImplementedError):
            BaseCalendarConverter().max_month(1900)
        with pytest.raises(NotImplementedError):
            BaseCalendarConverter().max_day(1900, 12)
        with pytest.raises(NotImplementedError):
            BaseCalendarConverter().to_gregorian(1900, 12, 31)
        with pytest.raises(NotImplementedError):
            BaseCalendarConverter().representative_years([1900, 1901])



from enum import IntEnum
from dataclasses import dataclass, replace
import operator

# Pre 3.10 requires Union for multiple types, e.g. Union[int, None] instead of int | None
from typing import Optional, Union, Iterable

import numpy as np


class Timedelta(np.ndarray):
    """Convenience class to make :class:`numpy.timedelta64` act
    more like the built-in python :class:`datetime.timedelta`."""

    def __new__(cls, deltadays: Union[np.timedelta64, int]):
        if isinstance(deltadays, int):
            deltadays = np.timedelta64(deltadays, "D")
        data = np.asarray(deltadays, dtype="timedelta64")
        return data.view(cls)

    def Export(self):
        return self

    def __array_finalize__(self, obj):
        if obj is None:
            return

    @property
    def days(self) -> int:
        """number of days, as an integer"""
        return int(self.astype("datetime64[D]").astype("int"))


@dataclass
class UnInt:
    """An uncertain integer intended for use with uncertain durations (:class:`UnDelta`),
    to convey a range of possible integer values between an upper
    and lower bound (both inclusive). Supports comparison, addition and subtraction,
    checking if a value is included in the range, and iterating over numbers
    included in the range.
    """

    lower: int
    upper: int

    def __post_init__(self):
        # validate that lower value is less than upper
        if not self.lower < self.upper:
            raise ValueError(
                f"Lower value ({self.lower}) must be less than upper ({self.upper})"
            )

    def __iter__(self) -> Iterable:
        # yield all integers in range from lower to upper, inclusive
        yield from range(self.lower, self.upper + 1)

    def __gt__(self, other: object) -> bool:
        match other:
            case int():
                return self.lower > other
            case UnInt():
                return self.lower > other.upper
            case _:
                return NotImplemented

    def __lt__(self, other: object) -> bool:
        match other:
            case int():
                return self.upper < other
            case UnInt():
                return self.upper < other.lower
            case _:
                return NotImplemented

    def __contains__(self, other: object) -> bool:
        match other:
            case int():
                return other >= self.lower and other <= self.upper
            case UnInt():
                return other.lower >= self.lower and other.upper <= self.upper
            case _:
                # unsupported type: return false
                return False

    def _replace_with(self, other_lower, other_upper, op):
        """Create and return a new instance of UnInt using the specified
        operator (e.g. add, subtract) and other values to modify the values in
        the current UnInt instance."""
        return replace(
            self, lower=op(self.lower, other_lower), upper=op(self.upper, other_upper)
        )

    def __add__(self, other: object) -> "UnInt":
        match other:
            case int():
                # increase both values by the added amount
                add_values = (other, other)
            case UnInt():
                # add other lower value to current lower and other upper
                # to current upper to include the largest range of possible values
                # (when calculating with uncertain values, the uncertainty increases)
                add_values = (other.lower, other.upper)
            case _:
                return NotImplemented

        return self._replace_with(*add_values, operator.add)

    def __sub__(self, other) -> "UnInt":
        match other:
            case int():
                # decrease both values by the subtracted amount
                sub_values = (other, other)
            case UnInt():
                # to determine the largest range of possible values,
                # subtract the other upper value from current lower
                # and other lower value from current upper
                sub_values = (other.upper, other.lower)
            case _:
                return NotImplemented

        return self._replace_with(*sub_values, operator.sub)


@dataclass
class UnDelta:
    """
    An uncertain timedelta, for durations where the number of days is uncertain.
    Initialize with a list of possible durations in days as integers, which are used
    to calculate a value for duration in :attr:`days` as an
    instance of :class:`UnInt`.
    """

    # NOTE: we will probably need other timedelta-like logic here besides days...

    #: possible durations days, as an instance of :class:`UnInt`
    days: UnInt

    def __init__(self, *days: int):
        if len(days) < 2:
            raise ValueError(
                "Must specify at least two values for an uncertain duration"
            )
        self.days = UnInt(min(days), max(days))

    def __repr__(self):
        # customize string representation  for simpler notation; default
        # specifies full UnInt initialization with upper and lower keywords
        return f"undate.{self.__class__.__name__}({self.days.lower},{self.days.upper})"

    def __eq__(self, other: object) -> bool:
        # is an uncertain duration ever *equal* another, even if the values are the same?
        # for now, make the assumption that we only want identity equality
        # and not value equality; perhaps in future we can revisit
        # or add functions to check value equality / equivalence / similarity
        return other is self

    def __lt__(self, other: object) -> bool:
        match other:
            case Timedelta() | UnDelta():
                return self.days < other.days
            case _:
                return NotImplemented

    def __gt__(self, other: object) -> bool:
        match other:
            case Timedelta() | UnDelta():
                return self.days > other.days
            case _:
                return NotImplemented


#: timedelta for single day
ONE_DAY = Timedelta(1)  # ~ equivalent to datetime.timedelta(days=1)
#: timedelta for a single  year (non-leap year)
ONE_YEAR = Timedelta(365)  # ~ relativedelta(years=1)
#: timedelta for a month, assuming maximum month length (31 days)
ONE_MONTH_MAX = Timedelta(31)


class Date(np.ndarray):
    """Convenience class to make :class:`numpy.datetime64` act
    more like the built-in python :class:`datetime.date`."""

    # extend np.datetime64 datatype
    # adapted from https://stackoverflow.com/a/27129510/9706217

    def __new__(
        cls,
        year: Union[int, np.datetime64],
        month: Optional[int] = None,
        day: Optional[int] = None,
    ):
        if isinstance(year, np.datetime64):
            _data = year
        else:
            datestr = str(year)
            if month is not None:
                datestr = f"{year}-{month:02d}"
                if day is not None:
                    datestr = f"{datestr}-{day:02d}"
            _data = np.datetime64(datestr)

        data = np.asarray(_data, dtype="datetime64")

        # expected dtype depends on date unit / how much of date is known
        expected_unit = "Y"
        if day is not None and month is not None:
            expected_unit = "D"
        elif month:
            expected_unit = "M"
        expected_dtype = f"datetime64[{expected_unit}]"

        if data.dtype != expected_dtype:
            raise ValueError(
                f"Unable to parse dates adequately as {expected_dtype}: {data}"
            )
        obj = data.view(cls)
        return obj

    def Export(self):
        return self

    def __array_finalize__(self, obj):
        if obj is None:
            return

    # custom properties to access year, month, day

    @property
    def year(self) -> int:
        return int(str(self.astype("datetime64[Y]")))

    @property
    def month(self) -> Optional[int]:
        # if date unit is year, don't return a month (only M/D)
        if self.dtype != "datetime64[Y]":
            return int(str(self.astype("datetime64[M]")).split("-")[-1])
        return None

    @property
    def day(self) -> Optional[int]:
        # only return a day if date unit is in days
        if self.dtype == "datetime64[D]":
            return int(str(self.astype("datetime64[D]")).split("-")[-1])
        return None

    @property
    def weekday(self) -> Optional[int]:
        """Equivalent to :meth:`datetime.date.weekday`; returns day of week as an
        integer where Monday is 0 and Sunday is 6. Only supported for dates
        with date unit in days.
        """
        # only return a weekday if date unit is in days
        if self.dtype == "datetime64[D]":
            # calculate based on difference between current day and week start
            # numpy datetime weeks start on thursdays - presumably since
            # unix epoch day zero was a thursday...

            # implementation inspired in part by https://stackoverflow.com/a/54264187

            thursday_week = self.astype("datetime64[W]")
            days_from_thursday = (self - thursday_week).astype(int)
            # if monday is 0, thursday is 3
            return (days_from_thursday + 3) % 7

        return None

    def __sub__(self, other):
        # modify to conditionally return a timedelta object instead of a
        # Date object with dtype timedelta64[D] (default behavior)

        result = super().__sub__(other)
        # if the result has a timedelta type (i.e., date minus date = timedelta),
        # cast to local Timedelta object; otherwise, leave as is
        # (i.e., date minus timedelta = date)
        if result.dtype == "timedelta64[D]":
            result = Timedelta(result)
        return result

    # NOTE: add should not be subclassed because we want to return a Date, not a delta


class DatePrecision(IntEnum):
    """date precision, to indicate date precision independent from how much
    of the date is known."""

    # NOTE: values MUST be ordered based on the relative size or
    # precison of the time unit. That is, the smaller the unit, the more precise
    # it is: a day is more precise than a month, a month is more precise than a year,
    # (DatePrecision.year < DatePrecision.month)

    #: century
    CENTURY = 1
    #: decade
    DECADE = 2
    #: year
    YEAR = 3
    # TBD: season ?
    #: month
    MONTH = 4
    #: day
    DAY = 5

    def __str__(self):
        return f"{self.name}"

    # NOTE: consider harmonizing / using numpy date units:
    # years (‘Y’), months (‘M’), weeks (‘W’), and days (‘D’)



from undate import Undate, UndateInterval
from undate.converters.iso8601 import ISO8601DateFormat


class TestISO8601DateFormat:
    def test_parse_singledate(self):
        assert ISO8601DateFormat().parse("2002") == Undate(2002)
        assert ISO8601DateFormat().parse("1991-05") == Undate(1991, 5)
        assert ISO8601DateFormat().parse("1991-05-03") == Undate(1991, 5, 3)
        # missing year but month/day known; compare repr string
        assert repr(ISO8601DateFormat().parse("--05-03")) == repr(
            Undate(month=5, day=3)
        )

    def test_parse_singledate_unequal(self):
        assert ISO8601DateFormat().parse("2002") != Undate(2003)
        assert ISO8601DateFormat().parse("1991-05") != Undate(1991, 6)
        assert ISO8601DateFormat().parse("1991-05-03") != Undate(1991, 5, 4)
        # missing year but month/day known
        assert ISO8601DateFormat().parse("--05-03") != Undate(month=5, day=4)

    def test_parse_invalid(self):
        # allow but warn?
        assert ISO8601DateFormat().parse("1991-5") == Undate(1991, 5)

    def test_parse_range(self):
        assert ISO8601DateFormat().parse("1800/1900") == UndateInterval(
            Undate(1800), Undate(1900)
        )

    def test_to_string(self):
        # NOTE: iso8601 to_string currently tested more thoroughly
        # in undate str tests; may want to move those tests here
        assert ISO8601DateFormat().to_string(Undate(900)) == "0900"
        assert ISO8601DateFormat().to_string(Undate(33)) == "0033"



import pytest

from undate.converters.calendars import IslamicDateConverter
from undate.converters.calendars.islamic.transformer import IslamicUndate
from undate.undate import Calendar, Undate
from undate.date import DatePrecision, Date


class TestIslamicDateConverter:
    def test_parse(self):
        # day
        # Monday, 7 Jumādā I 1243 Hijrī (26 November, 1827 CE); Jumada I = month 5
        date_str = "7 Jumādā I 1243"
        date = IslamicDateConverter().parse(date_str)
        assert date == IslamicUndate(1243, 5, 7)
        assert date.calendar == Calendar.ISLAMIC
        assert date.precision == DatePrecision.DAY
        assert date.label == f"{date_str} {IslamicDateConverter.calendar_name}"

        # month
        date_str = "Rajab 495"
        date = IslamicDateConverter().parse(date_str)
        assert date == IslamicUndate(495, 7)  # Rajab is month 7
        assert date.calendar == Calendar.ISLAMIC
        assert date.precision == DatePrecision.MONTH
        assert date.label == f"{date_str} {IslamicDateConverter.calendar_name}"
        # Gregorian earliest/ latest
        assert date.earliest == Date(1102, 4, 28)
        assert date.latest == Date(1102, 5, 27)

        # year
        date_str = "441"
        date = IslamicDateConverter().parse(date_str)
        assert date == IslamicUndate(441)
        assert date.calendar == Calendar.ISLAMIC
        assert date.precision == DatePrecision.YEAR
        assert date.label == f"{date_str} {IslamicDateConverter.calendar_name}"
        # Gregorian earliest/ latest
        assert date.earliest == Date(1049, 6, 11)
        assert date.latest == Date(1050, 5, 31)

    def test_gregorian_earliest_latest(self):
        # earliest/latest should be converted to Gregorian for comparison

        # Monday, 7 Jumādā I 1243 Hijrī (26 November, 1827 CE); Jumada I = month 5
        date = IslamicUndate(1243, 5, 7)
        assert date.earliest == Date(1827, 11, 26)
        assert date.latest == Date(1827, 11, 26)

        # Jumādā I 1243 : 1827-11-20 to 1827-12-19
        date = IslamicUndate(1243, 5)
        assert date.earliest == Date(1827, 11, 20)
        assert date.latest == Date(1827, 12, 19)

        # Rajab 495: 1102-04-28 to 1102-05-27 (Rajab = month 7)
        date = IslamicUndate(495, 7)
        assert date.earliest == Date(1102, 4, 28)
        assert date.latest == Date(1102, 5, 27)

        # 441 : 1049-06-11 to 1050-05-31
        date = IslamicUndate(441)
        assert date.earliest == Date(1049, 6, 11)
        assert date.latest == Date(1050, 5, 31)

        # examples from ISMI data (reformatted to day month year)
        # 14 Rabīʿ I 901 : 1495-12-11 (Rabi 1 = month 3 )
        date = IslamicUndate(901, 3, 14)
        assert date.earliest == Date(1495, 12, 11)
        assert date.latest == Date(1495, 12, 11)

        # 884 : 1479-04-03 to 1480-03-21
        date = IslamicUndate(884)
        assert date.earliest == Date(1479, 4, 3)
        assert date.latest == Date(1480, 3, 21)

    def test_parse_error(self):
        # a string we can't parse should raise an error
        with pytest.raises(ValueError):
            IslamicDateConverter().parse("January 2, 1991")
        # empty string should also error
        with pytest.raises(ValueError):
            IslamicDateConverter().parse("")

    def test_partially_known(self):
        # hijri dates get existing partially unknown behavior
        unknown_month = IslamicUndate(1243, "XX")
        assert unknown_month.precision == DatePrecision.MONTH
        assert unknown_month.earliest == Date(
            *IslamicDateConverter().to_gregorian(1243, 1, 1)
        )
        assert unknown_month.latest == Date(
            *IslamicDateConverter().to_gregorian(1243, 12, 30)
        )

        partially_unknown_month = IslamicUndate(1243, "1X")
        assert partially_unknown_month.precision == DatePrecision.MONTH
        assert partially_unknown_month.earliest == Date(
            *IslamicDateConverter().to_gregorian(1243, 10, 1)
        )
        assert partially_unknown_month.latest == Date(
            *IslamicDateConverter().to_gregorian(1243, 12, 30)
        )

        unknown_day = IslamicUndate(1243, 2, "XX")
        assert unknown_day.precision == DatePrecision.DAY
        assert unknown_day.earliest == Date(
            *IslamicDateConverter().to_gregorian(1243, 2, 1)
        )
        # second month has 29 days
        assert unknown_day.latest == Date(
            *IslamicDateConverter().to_gregorian(1243, 2, 29)
        )
        partially_unknown_day = IslamicUndate(1243, 2, "2X")
        assert partially_unknown_day.precision == DatePrecision.DAY
        assert partially_unknown_day.earliest == Date(
            *IslamicDateConverter().to_gregorian(1243, 2, 20)
        )
        assert partially_unknown_day.latest == Date(
            *IslamicDateConverter().to_gregorian(1243, 2, 29)
        )

    def test_compare_across_calendars(self):
        # only day-precision dates can be exactly equal across calendars

        # 7 Jumādā I 1243 Hijrī : 26 November, 1827; Jumada I = month 5
        assert IslamicUndate(1243, 5, 7) == Undate(1827, 11, 26)
        # 14 Rabīʿ I 901 : 1495-12-11 (Rabi 1 = month 3 )
        assert IslamicUndate(901, 3, 14) == Undate(1495, 12, 11)

        # greater than / less than
        assert IslamicUndate(901) < Undate(1500)
        assert IslamicUndate(901) > Undate(1450)
        # Jumādā I 1243 : 1827-11-20 to 1827-12-19
        assert IslamicUndate(1243, 5) > Undate(1827, 10)
        assert IslamicUndate(1243, 5) < Undate(1828, 1)

        # 7 Jumādā I 1243 Hijrī : 26 November, 1827, so it falls
        # within (or is contained by) November 1827
        assert IslamicUndate(1243, 5, 7) in Undate(1827, 11)
        assert IslamicUndate(1243, 5, 7) not in Undate(1827, 10)

        # sorting
        sorted_dates = sorted(
            [
                IslamicUndate(884),  # 1479 to 1480 Gregorian
                IslamicUndate(441),  # 1049 to 1050 Gregorian
                IslamicUndate(901),  # 1495 to 1495 Gregorian
                Undate(1995),
                Undate(33),
                Undate(1350),
            ]
        )
        expected_gregorian_years = [33, 1049, 1350, 1479, 1495, 1995]
        assert [d.earliest.year for d in sorted_dates] == expected_gregorian_years

    def test_representative_years(self):
        converter = IslamicDateConverter()
        # single year is not filtered
        # 1458 is a leap year; 1457 and 1459 are not
        assert converter.representative_years([1457]) == [1457]
        # multiple non-leap years, returns just the first
        assert converter.representative_years([1457, 1459]) == [1457]
        # next leap year is 2028; returns first leap year and first non-leap year, in input order
        assert converter.representative_years([1457, 1458, 1459]) == [1457, 1458]

        # if no years are provided, returns a known leap year and non-leap years
        assert converter.representative_years() == [
            converter.LEAP_YEAR,
            converter.NON_LEAP_YEAR,
        ]
        assert converter.representative_years([]) == [
            converter.LEAP_YEAR,
            converter.NON_LEAP_YEAR,
        ]



from undate.converters.calendars import SeleucidDateConverter
from undate.date import Date, DatePrecision
from undate.undate import Calendar, Undate


class TestSeleucidDateConverter:
    def test_parse(self):
        # day
        # Elul = month 6; 11 September, 1000 Gregorian
        date_str = "29 Elul 1311"
        date = SeleucidDateConverter().parse(date_str)
        assert date == Undate(1311, 6, 29, calendar="Seleucid")
        assert date.calendar == Calendar.SELEUCID
        assert date.precision == DatePrecision.DAY
        assert date.label == f"{date_str} {SeleucidDateConverter.calendar_name}"

        date_str = "Tishri 1458"  # month 7
        date = SeleucidDateConverter().parse(date_str)
        assert date == Undate(1458, 7, calendar="Seleucid")
        assert date.calendar == Calendar.SELEUCID
        assert date.precision == DatePrecision.MONTH
        assert date.label == f"{date_str} {SeleucidDateConverter.calendar_name}"

        # year
        date_str = "1458"
        date = SeleucidDateConverter().parse(date_str)
        assert date == Undate(1458, calendar="Seleucid")
        assert date.calendar == Calendar.SELEUCID
        assert date.precision == DatePrecision.YEAR
        assert date.label == f"{date_str} {SeleucidDateConverter.calendar_name}"

    def test_gregorian_earliest_latest(self):
        # earliest/latest should be converted to Gregorian for comparison

        # full date
        # Elul = month 6 (7 September, 1000 Gregorian)
        date_str = "29 Elul 1311"
        date = SeleucidDateConverter().parse(date_str)
        assert date.earliest == Date(1000, 9, 7)
        assert date.latest == Date(1000, 9, 7)
        assert date.label == f"{date_str} {SeleucidDateConverter.calendar_name}"

        date_str = "23 Adar I 1475"
        date = SeleucidDateConverter().parse(date_str)
        assert date.earliest == Date(1164, 2, 25)
        assert date.latest == Date(1164, 2, 25)
        assert date.label == f"{date_str} {SeleucidDateConverter.calendar_name}"

        # month/year
        date_str = "Tishri 1458"
        date = SeleucidDateConverter().parse(date_str)
        assert date.earliest == Date(1146, 9, 16)
        assert date.latest == Date(1146, 10, 15)
        assert date.label == f"{date_str} {SeleucidDateConverter.calendar_name}"

    def test_days_in_year(self):
        converter = SeleucidDateConverter()
        assert converter.days_in_year(2350) == 354
        assert converter.days_in_year(2349) == 385
        assert converter.days_in_year(2351) == 355


# TODO: update validation error to say seleucid instead of hebrew

#     seleucid_year = 1458
#     converted_date = convert_seleucid_date(f"Tishri {seleucid_year}")
#     converted_date_am = convert_hebrew_date(
#         f"Tishrei {seleucid_year + Calendar.SELEUCID_OFFSET}"
#     )
#     # the converted date range for Tishri Sel. should be the same as that for Tishri AM - 3449 years.
#     assert converted_date[0] == converted_date_am[0]
#     assert converted_date[1] == converted_date_am[1]

#     # leap day (Feb 29, 2020) should convert properlyd
#     converted_date = convert_seleucid_date("4 Adar 2331")
#     assert converted_date[1] == date(2020, 2, 29)


# # 26 Tammuz 4816: 17 July, 1056; Tammuz = month 4
# date = Undate(4816, 4, 26, calendar="Seleucid")
# assert date.earliest == Date(1056, 7, 17)
# assert date.latest == Date(1056, 7, 17)
# # 13 Tishrei 5416 Anno Mundi (1655-10-14)
# date = Undate(5416, 7, 13, calendar="Seleucid")  # Tishrei = month 7
# assert date.earliest == Date(1655, 10, 14)
# assert date.latest == Date(1655, 10, 14)


# from pgp tests


#     # month/year
#     seleucid_year = 1458
#     converted_date = convert_seleucid_date(f"Tishri {seleucid_year}")
#     converted_date_am = convert_hebrew_date(
#         f"Tishrei {seleucid_year + Calendar.SELEUCID_OFFSET}"
#     )
#     # the converted date range for Tishri Sel. should be the same as that for Tishri AM - 3449 years.
#     assert converted_date[0] == converted_date_am[0]
#     assert converted_date[1] == converted_date_am[1]

#     # leap day (Feb 29, 2020) should convert properly
#     converted_date = convert_seleucid_date("4 Adar 2331")
#     assert converted_date[1] == date(2020, 2, 29)

#     # leap year (4826 AM = 1377 Seleucid) should convert properly
#     seleucid_year = 1377
#     converted_date = convert_seleucid_date(f"21 Adar II {seleucid_year}")
#     converted_date_am = convert_hebrew_date(
#         f"21 Adar II {seleucid_year + Calendar.SELEUCID_OFFSET}"
#     )
#     assert converted_date[0] == converted_date_am[0]
#     assert converted_date[1] == converted_date_am[1]
#     # and it should be converted to 1066-03-21 CE
#     assert converted_date[1] == date(1066, 3, 21)



import datetime

import numpy as np
import pytest

from undate.date import (
    ONE_DAY,
    ONE_YEAR,
    ONE_MONTH_MAX,
    Date,
    DatePrecision,
    Timedelta,
    UnDelta,
    UnInt,
)


class TestDatePrecision:
    def test_str(self):
        assert str(DatePrecision.YEAR) == "YEAR"

    def test_precision_comparison(self):
        assert DatePrecision.DAY > DatePrecision.MONTH
        assert DatePrecision.MONTH > DatePrecision.YEAR


class TestDate:
    def test_init_year(self):
        d = Date(2001)
        assert isinstance(d, Date)
        assert d.dtype == "datetime64[Y]"
        assert str(d) == "2001"

    def test_init_year_np_datetime64(self):
        d = Date(np.datetime64("2024"))
        assert isinstance(d, Date)
        assert d.dtype == "datetime64[Y]"
        assert str(d) == "2024"

    def test_init_year_month(self):
        d = Date(2010, 5)
        assert isinstance(d, Date)
        assert d.dtype == "datetime64[M]"
        assert str(d) == "2010-05"

    def test_init_year_month_day(self):
        d = Date(2021, 6, 15)
        assert isinstance(d, Date)
        assert d.dtype == "datetime64[D]"
        assert str(d) == "2021-06-15"

    def test_properties_year(self):
        assert Date(2001).year == 2001
        assert Date(2010, 5).year == 2010
        assert Date(2021, 6, 15).year == 2021

    def test_properties_month(self):
        assert Date(2001).month is None
        assert Date(2010, 5).month == 5
        assert Date(2021, 6, 15).month == 6

    def test_properties_day(self):
        assert Date(2001).day is None
        assert Date(2010, 5).day is None
        assert Date(2021, 6, 15).day == 15

    def test_weekday(self):
        # thursday
        assert Date(2025, 1, 2).weekday == 3
        assert Date(2025, 1, 2).weekday == datetime.date(2025, 1, 2).weekday()
        # friday
        assert Date(2025, 1, 3).weekday == 4
        assert Date(2025, 1, 3).weekday == datetime.date(2025, 1, 3).weekday()
        # saturday
        assert Date(2025, 1, 4).weekday == 5
        assert Date(2025, 1, 4).weekday == datetime.date(2025, 1, 4).weekday()
        # sunday
        assert Date(2025, 1, 5).weekday == 6
        assert Date(2025, 1, 5).weekday == datetime.date(2025, 1, 5).weekday()
        # monday
        assert Date(2025, 1, 6).weekday == 0
        assert Date(2025, 1, 6).weekday == datetime.date(2025, 1, 6).weekday()
        # tuesday
        assert Date(2025, 1, 7).weekday == 1
        assert Date(2025, 1, 7).weekday == datetime.date(2025, 1, 7).weekday()

        # when a date is not day-level precision, no weekday is returned
        yearonly_date = Date(2025)
        assert yearonly_date.dtype == "datetime64[Y]"
        assert yearonly_date.weekday is None

    def test_substract(self):
        # date - date = timedelta
        date_difference = Date(2024, 1, 2) - Date(2024, 1, 1)
        assert isinstance(date_difference, Timedelta)
        assert date_difference.days == 1

        # date - timedelta = date
        year_prior = Date(2024, 1, 2) - ONE_YEAR
        assert isinstance(year_prior, Date)


class TestTimeDelta:
    def test_init_from_int(self):
        td = Timedelta(31)
        assert isinstance(td, Timedelta)
        assert td.dtype == "timedelta64[D]"
        assert td.astype("int") == 31

    def test_init_from_np_timedelta64(self):
        td = Timedelta(np.timedelta64(12, "D"))
        assert isinstance(td, Timedelta)
        assert td.dtype == "timedelta64[D]"
        assert td.astype("int") == 12

    def test_days(self):
        assert Timedelta(10).days == 10


class TestUnInt:
    def test_init(self):
        february_days = UnInt(28, 29)  # 28 or 29
        assert february_days.lower == 28
        assert february_days.upper == 29

        # also supports keyword args
        anymonth_days = UnInt(lower=28, upper=31)
        assert anymonth_days.lower == 28
        assert anymonth_days.upper == 31

    def test_init_validation(self):
        with pytest.raises(
            ValueError, match=r"Lower value \(10\) must be less than upper \(4\)"
        ):
            UnInt(10, 4)

    def test_contains(self):
        anymonth_days = UnInt(lower=28, upper=31)
        # integer
        assert 28 in anymonth_days
        assert 29 in anymonth_days
        assert 31 in anymonth_days
        assert 32 not in anymonth_days
        # unint
        assert UnInt(28, 29) in anymonth_days

        # other types are assumed not in range
        assert "twenty-eight" not in anymonth_days

    def test_gt(self):
        ten_twelve = UnInt(10, 12)
        # compare with integer
        assert ten_twelve > 9
        assert not ten_twelve > 12
        assert not ten_twelve > 15
        # compare with unint
        assert ten_twelve > UnInt(2, 4)
        assert not ten_twelve > UnInt(12, 24)
        assert not ten_twelve > UnInt(13, 23)
        # unsupported type
        with pytest.raises(TypeError):
            ten_twelve > "three"

    def test_lt(self):
        ten_twelve = UnInt(10, 12)
        # compare with integer
        assert ten_twelve < 13
        assert not ten_twelve < 12
        assert not ten_twelve < 9
        # compare with unint
        assert ten_twelve < UnInt(13, 23)
        assert not ten_twelve < UnInt(12, 24)
        assert not ten_twelve < UnInt(2, 4)
        # unsupported type
        with pytest.raises(TypeError):
            ten_twelve < "three"

    def test_iterable(self):
        anymonth_days = UnInt(lower=28, upper=31)
        assert list(anymonth_days) == [28, 29, 30, 31]

    def test_add(self):
        february_days = UnInt(28, 29)
        # add integer
        assert february_days + 1 == UnInt(29, 30)
        # add UnInt - minimum is 28 + 1, maximum is 29 + 2
        assert february_days + UnInt(1, 2) == UnInt(29, 31)
        # other types are not supported
        with pytest.raises(TypeError, match="unsupported operand"):
            february_days + "two"

    def test_subtract(self):
        february_days = UnInt(28, 29)
        # subtract integer
        assert february_days - 10 == UnInt(18, 19)
        # subtract UnInt - minimum is lower - largest value, maximum is upper - smallest value
        # difference between number of days in any month and the month of February?
        # [28,31] - [28,29] = [-1, 3]
        anymonth_days = UnInt(lower=28, upper=31)
        assert anymonth_days - february_days == UnInt(-1, 3)
        # what if we go the other direction?
        assert february_days - anymonth_days == UnInt(-3, 1)
        # other types are not supported
        with pytest.raises(TypeError, match="unsupported operand"):
            february_days - "two"


class TestUnDelta:
    def test_init(self):
        # February in an unknown year in Gregorian calendar could be 28 or 29 days
        february_days = UnInt(28, 29)  # 28 or 29
        udelt = UnDelta(28, 29)
        assert isinstance(udelt.days, UnInt)
        assert udelt.days.lower == 28
        assert udelt.days.upper == 29

        # NOTE: default portion interval comparison may not be what we want here,
        # since this is an unknown value within the range...
        # (maybe handled in undelta class comparison methods)
        assert udelt.days == february_days

        # do the right thing with more than one value, out of order
        unknown_month_duration = UnDelta(30, 31, 28)
        assert isinstance(unknown_month_duration.days, UnInt)
        assert unknown_month_duration.days.lower == 28
        assert unknown_month_duration.days.upper == 31

    def test_init_validation(self):
        with pytest.raises(ValueError, match="Must specify at least two values"):
            UnDelta(10)

    def test_repr(self):
        # test customized string representation

        # import undate to test eval of fully-qualified repr string
        import undate  # noqa: F401

        feb_undelt = UnDelta(28, 29)
        assert repr(feb_undelt) == "undate.UnDelta(28,29)"
        # can't compare directly because uncertain deltas aren't equal,
        # but compare values
        assert eval(repr(feb_undelt.days.lower)) == feb_undelt.days.lower
        assert eval(repr(feb_undelt.days.upper)) == feb_undelt.days.upper

        larger_undelt = UnDelta(10, 12, 14, 16)
        assert repr(larger_undelt) == "undate.UnDelta(10,16)"

    def test_eq(self):
        # uncertain deltas are not equivalent
        udelt1 = UnDelta(30, 31)
        udelt2 = UnDelta(30, 31)
        # not equal to equivalent undelta range
        assert udelt1 != udelt2
        # equal to self
        assert udelt1 is udelt1

    def test_lt(self):
        week_or_tenday = UnDelta(7, 10)
        # compare undelta with undelta
        month = UnDelta(28, 31)
        # a week or ten-day is unambiguously less than a month
        assert week_or_tenday < month
        # compare undelta with Timedelta
        # NOTE: currently requires this direction, until we update Timedelta
        assert not week_or_tenday < ONE_DAY
        # an uncertain  month is unambiguously less than a year
        assert month < ONE_YEAR
        # an uncertain  month may or may not be less than one month max
        assert not month < ONE_MONTH_MAX

    def test_gt(self):
        week_or_tenday = UnDelta(7, 10)
        # compare undelta with undelta
        month = UnDelta(28, 31)
        # a month is unambiguously longer than week or ten-day
        assert month > week_or_tenday
        # compare undelta with Timedelta
        # NOTE: currently requires this direction, until we update Timedelta
        # to support the reverse comparison
        assert week_or_tenday > ONE_DAY
        # an uncertain month is not greater than a year
        assert not month > ONE_YEAR
        # an uncertain  month may or may not be greater than one month max
        assert not month > ONE_MONTH_MAX



import pytest
from undate.converters.edtf import EDTFDateConverter
from undate import Undate, UndateInterval


class TestEDTFDateConverter:
    def test_parse_singledate(self):
        assert EDTFDateConverter().parse("2002") == Undate(2002)
        assert EDTFDateConverter().parse("1991-05") == Undate(1991, 5)
        assert EDTFDateConverter().parse("1991-05-03") == Undate(1991, 5, 3)
        # unknown dates are not strictly equal, but string comparison should match
        assert str(EDTFDateConverter().parse("201X")) == str(Undate("201X"))
        assert str(EDTFDateConverter().parse("2004-XX")) == str(Undate(2004, "XX"))
        assert str(EDTFDateConverter().parse("XXXX-05-03")) == Undate(
            month=5, day=3
        ).format("EDTF")

    def test_parse_singledate_error(self):
        # missing year but month/day known - this format should cause an error
        with pytest.raises(ValueError):
            EDTFDateConverter().parse("--05-03")

    def test_parse_singledate_unequal(self):
        assert EDTFDateConverter().parse("2002") != Undate(2003)
        assert EDTFDateConverter().parse("1991-05") != Undate(1991, 6)
        assert EDTFDateConverter().parse("1991-05-03") != Undate(1991, 5, 4)
        # missing year but month/day known
        # - does EDTF not support this or is parsing logic incorrect?
        # assert EDTFDateConverter().parse("XXXX-05-03") != Undate(month=5, day=4)

    invalid_inputs = [
        "1991-13",  # invalid month
        "1991-12-32",  # invalid day
        "199A",  # invalid year format
        "",  # empty string
        None,  # None input
    ]

    @pytest.mark.parametrize("invalid_input", invalid_inputs)
    def test_parse_invalid(self, invalid_input):
        with pytest.raises(ValueError):
            EDTFDateConverter().parse(invalid_input)

    def test_parse_range(self):
        assert EDTFDateConverter().parse("1800/1900") == UndateInterval(
            Undate(1800), Undate(1900)
        )

    def test_to_string(self):
        assert EDTFDateConverter().to_string(Undate(900)) == "0900"
        assert EDTFDateConverter().to_string(Undate("80")) == "0080"
        assert EDTFDateConverter().to_string(Undate(33)) == "0033"
        assert EDTFDateConverter().to_string(Undate("20XX")) == "20XX"
        assert EDTFDateConverter().to_string(Undate(17000002)) == "Y17000002"
        assert EDTFDateConverter().to_string(Undate(-17000002)) == "Y-17000002"

        assert EDTFDateConverter().to_string(Undate(1991, 6)) == "1991-06"
        assert EDTFDateConverter().to_string(Undate(1991, 5, 3)) == "1991-05-03"

        assert EDTFDateConverter().to_string(Undate(1991, "0X")) == "1991-0X"
        assert EDTFDateConverter().to_string(Undate(1991, None, 3)) == "1991-XX-03"
        assert EDTFDateConverter().to_string(Undate(-1984)) == "-1984"

        # if converter can't generate a string for the date,
        # it should return a value error
        # empty_undate = Undate()   # undate with no date information no longer supported
        # empty_undate.precision = DatePrecision.DECADE
        # with pytest.raises(ValueError):
        #     EDTFDateConverter().to_string(empty_undate)
        # TODO: override missing digit and confirm replacement



from __future__ import annotations

import datetime
from enum import auto

import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from undate.interval import UndateInterval

try:
    # StrEnum was only added in python 3.11
    from enum import StrEnum
except ImportError:
    # for python 3.10 or earlier, use third-party package
    from strenum import StrEnum  # type: ignore

# Pre 3.10 requires Union for multiple types, e.g. Union[int, None] instead of int | None
from typing import Dict, Optional, Union

from undate.converters.base import BaseCalendarConverter, BaseDateConverter
from undate.date import ONE_DAY, Date, DatePrecision, Timedelta, UnDelta


class Calendar(StrEnum):
    """Supported calendars"""

    GREGORIAN = auto()
    HEBREW = auto()
    ISLAMIC = auto()
    SELEUCID = auto()

    @staticmethod
    def get_converter(calendar) -> BaseCalendarConverter:
        # calendar converter must be available with a name matching
        # the title-case name of the calendar enum entry
        try:
            converter_cls = BaseDateConverter.available_converters()[
                calendar.value.title()
            ]
        except KeyError as err:
            raise ValueError(f"Unknown calendar '{calendar}'") from err
        if not issubclass(converter_cls, BaseCalendarConverter):
            raise ValueError(
                f"Requested converter '{calendar.value.title()}' is not a CalendarConverter"
            )
        return converter_cls()


class Undate:
    """object for representing uncertain, fuzzy or partially unknown dates"""

    DEFAULT_CONVERTER: str = "ISO8601"

    #: symbol for unknown digits within a date value
    MISSING_DIGIT: str = "X"

    earliest: Date
    latest: Date
    #: A string to label a specific undate, e.g. "German Unity Date 2022" for Oct. 3, 2022.
    #: Labels are not taken into account when comparing undate objects.
    label: Union[str, None] = None
    converter: BaseDateConverter
    #: precision of the date (day, month, year, etc.)
    precision: DatePrecision
    #: the calendar this date is using; Gregorian by default
    calendar: Calendar = Calendar.GREGORIAN

    # numpy datetime is stored as 64-bit integer, so min/max
    # depends on the time unit; assume days for now
    # See https://numpy.org/doc/stable/reference/arrays.datetime.html#datetime-units
    # It just so happens that int(2.5e16) is a leap year, which is a weird default,
    # so let's increase our lower bound by one year.
    MIN_ALLOWABLE_YEAR = int(-2.5e16) + 1
    MAX_ALLOWABLE_YEAR = int(2.5e16)

    def __init__(
        self,
        year: Optional[Union[int, str]] = None,
        month: Optional[Union[int, str]] = None,
        day: Optional[Union[int, str]] = None,
        converter: Optional[BaseDateConverter] = None,
        label: Optional[str] = None,
        calendar: Optional[Union[str, Calendar]] = None,
    ):
        # everything is optional but something is required
        if all([val is None for val in [year, month, day]]):
            raise ValueError("At least one of year, month, or day must be specified")

        # keep track of initial values and which values are known
        # TODO: add validation: if str, must be expected length
        self.initial_values: Dict[str, Optional[Union[int, str]]] = {
            "year": year,
            "month": month,
            "day": day,
        }
        if day:
            self.precision = DatePrecision.DAY
        elif month:
            self.precision = DatePrecision.MONTH
        elif year:
            self.precision = DatePrecision.YEAR

        self.label = label
        if calendar is not None:
            self.set_calendar(calendar)
        self.calendar_converter = Calendar.get_converter(self.calendar)
        self.calculate_earliest_latest(year, month, day)

        if converter is None:
            #  import all subclass definitions; initialize the default
            converter_cls = BaseDateConverter.available_converters()[
                self.DEFAULT_CONVERTER
            ]
            converter = converter_cls()
        self.converter = converter

    def calculate_earliest_latest(self, year, month, day):
        # special case: treat year = XXXX as unknown/none
        if year == "XXXX":
            year = None
        if year is not None:
            # could we / should we use str.isnumeric here?
            try:
                year = int(year)
                # update initial value since it is used to determine
                # whether or not year is known
                self.initial_values["year"] = year
                min_year = max_year = year
            except ValueError:
                # year is a string that can't be converted to int
                min_year = int(str(year).replace(self.MISSING_DIGIT, "0"))
                max_year = int(str(year).replace(self.MISSING_DIGIT, "9"))
        else:
            # if we don't have any other bounds,
            # use calendar-specific min year if there is one, otherwise use
            # the configured min/max allowable years
            min_year = self.calendar_converter.MIN_YEAR or self.MIN_ALLOWABLE_YEAR
            max_year = self.calendar_converter.MAX_YEAR or self.MAX_ALLOWABLE_YEAR

        # if month is passed in as a string but completely unknown,
        # treat as unknown/none (date precision already set in init)
        if month == "XX":
            month = None

        # get first and last month from the calendar (not always 1 and 12)
        # as well as min/max months
        earliest_month = self.calendar_converter.first_month()
        latest_month = self.calendar_converter.last_month(max_year)

        min_month = self.calendar_converter.min_month()
        max_month = self.calendar_converter.max_month(max_year)
        if month is not None:
            try:
                # treat as an integer if we can
                month = int(month)
                # update initial value
                self.initial_values["month"] = month
                earliest_month = latest_month = month
            except ValueError:
                # if not, calculate min/max for missing digits
                earliest_month, latest_month = self._missing_digit_minmax(
                    str(month), min_month, max_month
                )
        # similar to month above — unknown day, but day-level granularity
        if day == "XX":
            day = None

        # if day is numeric, use as is
        if isinstance(day, int) or isinstance(day, str) and day.isnumeric():
            day = int(day)
            # update initial value - fully known day
            self.initial_values["day"] = day
            min_day = max_day = day
        else:
            # if we have no day or partial day, calculate min / max
            min_day = 1  # is min day ever anything other than 1 ?
            rel_year = year if year and isinstance(year, int) else max_year
            # use month if it is an integer; otherwise use previusly determined
            # max month (which may not be 12 depending if partially unknown)
            rel_month = month if month and isinstance(month, int) else latest_month

            max_day = self.calendar_converter.max_day(rel_year, rel_month)

            # if day is partially specified, narrow min/max further
            if day is not None:
                min_day, max_day = self._missing_digit_minmax(day, min_day, max_day)

        # TODO: special case, if we get a Feb 29 date with unknown year,
        # should switch the min/max years to known leap years!

        # for unknowns, assume smallest possible value for earliest and
        # largest valid for latest
        # convert to Gregorian calendar so earliest/latest can always
        # be used for comparison
        self.earliest = Date(
            *self.calendar_converter.to_gregorian(min_year, earliest_month, min_day)
        )
        self.latest = Date(
            *self.calendar_converter.to_gregorian(max_year, latest_month, max_day)
        )

    def set_calendar(self, calendar: Union[str, Calendar]):
        """Find calendar by name if passed as string and set on the object.
        Only intended for use at initialization time; use :meth:`as_calendar`
        to change calendar."""
        if calendar is not None:
            # if not passed as a Calendar instance, do a lookup
            if isinstance(calendar, str):
                # look for calendar by upper-case name
                try:
                    calendar = Calendar[calendar.upper()]
                except KeyError as err:
                    raise ValueError(f"Calendar `{calendar}` is not supported") from err
            self.calendar = calendar

    def as_calendar(self, calendar: Union[str, Calendar]):
        """Return a new :class:`Undate` object with the same year, month, day, and labels
        used to initialize the current object, but with a different calendar.  Note that this
        does NOT do calendar conversion, but reinterprets current numeric year, month, day values
        according to the new calendar."""
        return Undate(
            year=self.initial_values.get("year"),
            month=self.initial_values.get("month"),
            day=self.initial_values.get("day"),
            label=self.label,
            calendar=calendar,
        )

    def __str__(self) -> str:
        # if any portion of the date is partially known, construct
        # pseudo ISO8601 format here, since ISO8601 doesn't support unknown digits
        # (temporary, should switch to default format that can handle it, e.g. EDTF)
        if any(self.is_partially_known(part) for part in ["year", "month", "day"]):
            # initial values could be either string or int
            year = self.initial_values["year"]
            month = self.initial_values["month"]
            day = self.initial_values["day"]
            # if integer, convert to string with correct number of digits
            # replace unknown year with - for --MM or --MM-DD format
            parts = [
                f"{year:04d}" if isinstance(year, int) else year or "-",
                f"{month:02d}" if isinstance(month, int) else month,
                f"{day:02d}" if isinstance(day, int) else day,
            ]
            # combine, skipping any values that are None
            return "-".join([str(p) for p in parts if p is not None])

        return self.converter.to_string(self)

    def __repr__(self) -> str:
        init_opts = {k: v for k, v in self.initial_values.items() if v is not None}
        # include label if set
        if self.label:
            init_opts["label"] = self.label
        # always include calendar
        init_opts["calendar"] = self.calendar.value.title()
        # combine parameters; use !r to quote strings
        init_str = ", ".join([f"{key}={val!r}" for key, val in init_opts.items()])
        return f"undate.Undate({init_str})"

    @classmethod
    def parse(cls, date_string, format) -> Union["Undate", UndateInterval]:
        """parse a string to an undate or undate interval using the specified format;
        for now, only supports named converters"""
        converter_cls = BaseDateConverter.available_converters().get(format, None)
        if converter_cls:
            # NOTE: some parsers may return intervals; is that ok here?
            return converter_cls().parse(date_string)

        raise ValueError(f"Unsupported format '{format}'")

    def format(self, format) -> str:
        """format this undate as a string using the specified format;
        for now, only supports named converters"""
        converter_cls = BaseDateConverter.available_converters().get(format, None)
        if converter_cls is not None:
            # NOTE: some parsers may return intervals; is that ok here?
            return converter_cls().to_string(self)

        raise ValueError(f"Unsupported format '{format}'")

    @classmethod
    def _comparison_type(cls, other: object) -> "Undate":
        """Common logic for type handling in comparison methods.
        Converts to Undate object if possible, otherwise raises
        NotImplementedError exception.  Uses :meth:`to_undate` for conversion.
        """
        # convert if possible; return NotImplemented if not
        try:
            return cls.to_undate(other)
        except TypeError:
            # recommended to support comparison with arbitrary objects
            return NotImplemented

    def __eq__(self, other: object) -> bool:
        # Note: assumes label differences don't matter for comparing dates

        # only a day-precision fully known undate can be equal to a datetime.date
        if isinstance(other, datetime.date):
            return self.earliest == other and self.latest == other

        other = self._comparison_type(other)
        if other is NotImplemented:
            # return NotImplemented to indicate comparison is not supported
            # with this type
            return NotImplemented

        # if either date has an unknown year, then not equal
        if not self.known_year or not other.known_year:
            return False

        # if both dates are fully known, then earliest/latest check
        # is sufficient (and will work across calendars!)

        # check for apparent equality
        # - earliest/latest match and both have the same precision
        looks_equal = (
            self.earliest == other.earliest
            and self.latest == other.latest
            and self.precision == other.precision
        )
        # if everything looks the same, check for any unknowns in initial values
        # the same unknown date should NOT be considered equal
        # (but do we need a different equivalence check for this?)

        # NOTE: assumes that partially known values can only be written
        # in one format (i.e. X for missing digits).
        # If we support other formats, will need to normalize to common
        # internal format for comparison
        if looks_equal and (
            # if any part of either date that is known is _partially_ known,
            # then these dates are not equal
            any([self.is_partially_known(p) for p in self.initial_values.keys()])
            or any([other.is_partially_known(p) for p in other.initial_values.keys()])
        ):
            return False

        return looks_equal

    def __lt__(self, other: object) -> bool:
        other = self._comparison_type(other)
        if other is NotImplemented:
            # return NotImplemented to indicate comparison is not supported
            # with this type
            return NotImplemented

        # if either date has a completely unknown year, then we can't compare
        if self.unknown_year or other.unknown_year:
            return False

        # if this date ends before the other date starts,
        # return true (this date is earlier, so it is less)
        if self.latest < other.earliest:
            return True

        # if the other one ends before this one starts,
        # return false (this date is later, so it is not less)
        if other.latest < self.earliest:
            return False

        # if it does not, check if one is included within the other
        # (e.g., single date within the same year)
        # comparison for those cases is not currently supported
        elif other in self or self in other:
            # sort by precision, most precise first
            by_precision = sorted(
                [self, other], key=lambda x: x.precision, reverse=True
            )
            raise NotImplementedError(
                f"Can't compare when one date ({by_precision[0]}) falls within the other ({by_precision[1]})"
            )
        # NOTE: unsupported comparisons are supposed to return NotImplemented
        # However, doing that in this case results in a confusing TypeError!
        #   TypeError: '<' not supported between instances of 'Undate' and 'Undate'
        # How to handle when the comparison is ambiguous / indeterminate?
        # we may need a tribool / ternary type (true, false, unknown),
        # but not sure what python builtin methods will do with it (unknown = false?)

        # for any other case (i.e., self == other), return false
        return False

    def __gt__(self, other: object) -> bool:
        # define gt ourselves so we can support > comparison with datetime.date,
        # but rely on existing less than implementation.
        # strictly greater than must rule out equals

        # if either date has a completely unknown year, then we can't compare
        # NOTE: this means that gt and lt will both be false when comparing
        # with a date with an unknown year...
        if self.unknown_year or isinstance(other, Undate) and other.unknown_year:
            return False

        return not (self < other or self == other)

    def __le__(self, other: object) -> bool:
        # if either date has a completely unknown year, then we can't compare
        if self.unknown_year or isinstance(other, Undate) and other.unknown_year:
            return False

        return self == other or self < other

    def __contains__(self, other: object) -> bool:
        # if the two dates are strictly equal, don't consider
        # either one as containing the other
        other = self._comparison_type(other)
        if other is NotImplemented:
            # return NotImplemented to indicate comparison is not supported
            # with this type
            return NotImplemented

        if self == other:
            return False

        # if either date has a completely unknown year, then we can't determine
        if self.unknown_year or other.unknown_year:
            return False

        return all(
            [
                self.earliest <= other.earliest,
                self.latest >= other.latest,
                # is precision sufficient for comparing partially known dates?
                # checking based on less precise /less granular time unit,
                # e.g. a day or month could be contained in a year
                # but not the reverse
                self.precision < other.precision,
            ]
        )

    @classmethod
    def to_undate(cls, other: object) -> "Undate":
        """Convert arbitrary object to Undate, if possible. Raises TypeError
        if conversion is not possible.

        Currently supports:
            - :class:`datetime.date` or :class:`datetime.datetime`
            - :class:`undate.date.Date`

        """
        match other:
            case Undate():
                return other
            case datetime.date() | datetime.datetime():
                return Undate(other.year, other.month, other.day)
            case Date():
                # handle conversion from internal Date class
                return Undate(other.year, other.month, other.day)

            case _:
                raise TypeError(f"Conversion from {type(other)} is not supported")

    @property
    def known_year(self) -> bool:
        "year is fully known"
        return self.is_known("year")

    @property
    def unknown_year(self) -> bool:
        "year is completely unknown"
        return self.is_unknown("year")

    def is_known(self, part: str) -> bool:
        """Check if a part of the date (year, month, day) is fully known.
        Returns False if unknown or only partially known."""
        # TODO: should we use constants or enum for values?

        # if we have an integer, then consider the date known
        # if we have a string, then it is only partially known; return false
        return isinstance(self.initial_values[part], int)

    def is_unknown(self, part: str) -> bool:
        """Check if a part of the date (year, month, day) is completely unknown."""
        return self.initial_values.get(part) is None

    def is_partially_known(self, part: str) -> bool:
        # TODO: should XX / XXXX really be considered partially known?
        # other code seems to assume this, so we'll preserve the behavior
        return isinstance(self.initial_values[part], str)
        # and self.initial_values[part].replace(self.MISSING_DIGIT, "") != ""

    @property
    def year(self) -> Optional[str]:
        "year as string (minimum 4 characters), if year is known"
        year = self._get_date_part("year")
        if year:
            return f"{year:0>4}"
        # if value is unset but date precision is year or greater, return unknown year
        elif self.precision >= DatePrecision.YEAR:
            return self.MISSING_DIGIT * 4
        return None

    @property
    def month(self) -> Optional[str]:
        "month as 2-character string, or None if unknown/unset"
        # TODO: do we allow None for unknown month with day-level granularity?
        # TODO: need to distinguish between unknown (XX) and unset/not part of the date due to granularity
        month = self._get_date_part("month")
        if month:
            return f"{month:0>2}"
        # if value is unset but date precision is month or greater, return unknown month
        elif self.precision >= DatePrecision.MONTH:
            return self.MISSING_DIGIT * 2
        return None

    @property
    def day(self) -> Optional[str]:
        "day as 2-character string or None if unset"
        day = self._get_date_part("day")
        if day:
            return f"{day:0>2}"
        # if value is unset but date precision is day, return unknown day
        # (may not be possible to have day precision with day part set in normal use)
        elif self.precision == DatePrecision.DAY:
            return self.MISSING_DIGIT * 2
        return None

    def _get_date_part(self, part: str) -> Optional[str]:
        value = self.initial_values.get(part)
        return str(value) if value else None

    @property
    def possible_years(self) -> list[int] | range:
        """A list or range of possible years for this date in the original calendar.
        Returns a list with a single year for dates with fully-known years."""
        # get the initial value passed in for year in original calendar
        initial_year_value = self.initial_values["year"]
        # if integer, year is fully known and is the only possible value
        if isinstance(initial_year_value, int):
            return [initial_year_value]

        # if year is None or string with all unknown digits, bail out
        if (
            initial_year_value is None
            or str(self.year).replace(self.MISSING_DIGIT, "") == ""
        ):
            # otherwise, year is fully unknown
            # returning range from min year to max year is not useful in any scenario!
            raise ValueError(
                "Possible years cannot be returned for completely unknown year"
            )

        # otherwise, year is partially known
        # determine the smallest step size for the missing digit
        earliest_year = int(str(self.year).replace(self.MISSING_DIGIT, "0"))
        latest_year = int(str(self.year).replace(self.MISSING_DIGIT, "9"))
        missing_digit_place = len(str(self.year)) - str(self.year).rfind(
            self.MISSING_DIGIT
        )
        # convert place to 1, 10, 100, 1000, etc.
        step = 10 ** (missing_digit_place - 1)
        # generate a range from earliest to latest with the appropriate step
        # based on the smallest missing digit
        return range(earliest_year, latest_year + 1, step)

    @property
    def representative_years(self) -> list[int]:
        """A list of representative years for this date."""
        try:
            # todo: filter by calendar to minimum needed
            try:
                return self.calendar_converter.representative_years(
                    list(self.possible_years)
                )
            except NotImplementedError:
                # if calendar converter does not support representative years, return all years
                return list(self.possible_years)
        except ValueError:
            return [
                self.calendar_converter.LEAP_YEAR,
                self.calendar_converter.NON_LEAP_YEAR,
            ]

    def duration(self) -> Timedelta | UnDelta:
        """What is the duration of this date?
        Calculate based on earliest and latest date within range,
        taking into account the precision of the date even if not all
        parts of the date are known. Note that durations are inclusive
        (i.e., a closed interval)  and include both the earliest and latest
        date rather than the difference between them.  Returns a :class:`undate.date.Timedelta` when
        possible, and an :class:`undate.date.UnDelta` when the duration is uncertain."""

        # if precision is a single day, duration is one day
        # no matter when it is or what else is known
        if self.precision == DatePrecision.DAY:
            return ONE_DAY

        # if year is known and no values are partially known,
        # we can calculate a time delta based on earliest + latest
        if self.known_year and not any(
            [self.is_partially_known(part) for part in ["year", "month", "day"]]
        ):
            #  subtract earliest from latest and add a day to include start day in the count
            return self.latest - self.earliest + ONE_DAY

        possible_max_days = set()
        # if precision is month and year is unknown,
        # calculate month duration within a single year (not min/max)
        if self.precision == DatePrecision.MONTH:
            # for every possible month and year, get max days for that month,
            # appease mypy, which says month values could be None here;
            # Date object allows optional month, but earliest/latest initialization
            # should always be day-precision dates

            # use months from the original calendar, not months from
            # earliest/latest dates, which have been converted to Gregorian
            initial_month_value = self.initial_values["month"]
            # if integer, month is fully known and is the only possible value
            possible_months: list[int] | range
            if isinstance(initial_month_value, int):
                possible_months = [initial_month_value]
            elif isinstance(initial_month_value, str):
                # determine earliest and latest possible months
                # based on missing digits and calendar
                year = (
                    self.year
                    if isinstance(self.year, int)
                    else self.calendar_converter.LEAP_YEAR
                )
                earliest_month, latest_month = self._missing_digit_minmax(
                    initial_month_value,
                    self.calendar_converter.min_month(),
                    self.calendar_converter.max_month(year),
                )
                possible_months = range(earliest_month, latest_month + 1)

            for month in possible_months:
                for year in self.representative_years:
                    possible_max_days.add(self.calendar_converter.max_day(year, month))

        # if precision is year but year is unknown, return an uncertain delta
        elif self.precision == DatePrecision.YEAR:
            # this is currently hebrew-specific due to the way the start/end of year wraps for that calendar
            # with contextlib.suppress(NotImplementedError):
            possible_max_days = {
                self.calendar_converter.days_in_year(y)
                for y in self.representative_years
            }

        # if there is more than one possible value for number of days
        # due to range including lear year / non-leap year, return an uncertain delta
        if len(possible_max_days) > 1:
            return UnDelta(*possible_max_days)
        return Timedelta(possible_max_days.pop())

    def _missing_digit_minmax(
        self, value: str, min_val: int, max_val: int
    ) -> tuple[int, int]:
        # given a possible range, calculate min/max values for a string
        # with a missing digit

        # TODO: test this method directly

        # assuming two digit only (i.e., month or day)
        possible_values = [f"{n:02}" for n in range(min_val, max_val + 1)]
        # ensure input value has two digits
        value = "%02s" % value
        # generate regex where missing digit matches anything
        val_pattern = re.compile(value.replace(self.MISSING_DIGIT, "."))
        # identify all possible matches, then get min and max
        matches = [val for val in possible_values if val_pattern.match(val)]
        min_match = min(matches)
        max_match = max(matches)

        # split input string into a list so we can update individually
        new_min_val = list(value)
        new_max_val = list(value)
        for i, digit in enumerate(value):
            # replace the corresponding digit with our min and max
            if digit == self.MISSING_DIGIT:
                new_min_val[i] = min_match[i]
                new_max_val[i] = max_match[i]

        # combine the lists of digits back together and convert to int
        min_val = int("".join(new_min_val))
        max_val = int("".join(new_max_val))
        return (min_val, max_val)



import pytest
from lark.exceptions import LarkError

from undate.converters.calendars.islamic.parser import islamic_parser


# for now, just test that valid dates can be parsed

testcases = [
    # year
    "521",
    # month + year
    # - with and without accent
    "al-Muḥarram 900",
    "al-Muharram 900",
    "Safar 581",
    "Ṣafar 581",
    # variant month name, with or without accent
    "Muharram 900",
    "Muḥarram 900",
    "Rabīʿ al-ʾAwwal 901",
    "Rabi' I 901",
    "Rabīʿ ath-Thānī 343",
    "Rabīʿ II 343",
    "Jumādā al-ʾAwwal 1081",
    "Jumada al-`Awwal 1081",
    "Jumādā I 1081",
    "Jumādā ath-Thāniyah 901",
    "Jumada ath-Thaniyah 901",
    "Jumādā II 981",
    "Rajab 942",
    "Shaʿbān 900",
    "Sha'ban 900",
    "Ramaḍān 903",
    "Ramadan 903",
    "Shawwāl 1042",
    "Shawwal 1042",
    "Zū al-Qaʿdah 124",
    "Dhu l-Qa'da 124",
    # day month year
    "7 Jumādā I 1243",
    "29 Muḥarram 1243",
    "30 Muḥarram 1243",
    "Rabīʿ I 901",
    "12 Rabīʿ I 901",
    # two and 1 digit years
    "12 Rabīʿ I 90",
    "12 Rabīʿ I 9",
]


@pytest.mark.parametrize("date_string", testcases)
def test_should_parse(date_string):
    assert islamic_parser.parse(date_string)


error_cases = [
    # invalid days
    "0 Muḥarram 1243",
    "31 Muḥarram 1243",
    # month alone
    "Shawwal",
    # month day only
    "12 Shawwal",
    # invalid month
    "Foo 383",
    # wrong format
    "2024-10-02",
    # year month day not supported
    "901 Rabīʿ I",
    "901 Rabīʿ I 12",
]


@pytest.mark.parametrize("date_string", error_cases)
def test_should_error(date_string):
    with pytest.raises(LarkError):
        islamic_parser.parse(date_string)



import pytest
from undate.converters.calendars.hebrew.parser import hebrew_parser
from undate.converters.calendars.hebrew.transformer import (
    HebrewDateTransformer,
    HebrewUndate,
)
from undate.undate import Undate, Calendar
from undate.date import DatePrecision


def test_hebrew_undate():
    assert HebrewUndate(848).calendar == Calendar.HEBREW


testcases = [
    # examples from Princeton Geniza Project
    # date conversions checked with https://www.muqawwim.com/
    # 26 Tammuz 4816; Tammuz  = month 4
    ("26 Tammuz 4816", HebrewUndate(4816, 4, 26), DatePrecision.DAY),
    ("Tammuz 4816", HebrewUndate(4816, 4), DatePrecision.MONTH),
    ("4816", HebrewUndate(4816), DatePrecision.YEAR),
    # 26 Tishrei 5416: Tishrei = month 7
    ("26 Tishrei 5416", HebrewUndate(5416, 7, 26), DatePrecision.DAY),
    # Ṭeveth = month 10
    ("Ṭevet 5362", HebrewUndate(5362, 10), DatePrecision.MONTH),
    ("5362", HebrewUndate(5362), DatePrecision.YEAR),
    # add when we support parsing ranges:
    # Adar I and Adar II 5453 : (1693 CE)
    # support weekdays included in text
    ("Thursday, 12 Sivan 4795", HebrewUndate(4795, 3, 12), DatePrecision.DAY),
    # with or without comma
    ("Thursday 12 Sivan 4795", HebrewUndate(4795, 3, 12), DatePrecision.DAY),
    # huh, current parsing completely ignores whitespace; do we want that?
    ("Thursday12Sivan4795", HebrewUndate(4795, 3, 12), DatePrecision.DAY),
]


@pytest.mark.parametrize("date_string,expected,expected_precision", testcases)
def test_transform(date_string, expected, expected_precision):
    transformer = HebrewDateTransformer(visit_tokens=True)
    # parse the input string, then transform to undate object
    parsetree = hebrew_parser.parse(date_string)
    transformed_date = transformer.transform(parsetree)
    assert transformed_date == expected
    # currently only undates have date precision
    if isinstance(transformed_date, Undate):
        assert transformed_date.precision == expected_precision
    # transformer doesn't have access to date string,
    # label will need to be set by the converter class



from datetime import date, datetime
from enum import auto
from unittest import mock

import pytest

from undate import Undate, UndateInterval, Calendar
from undate.undate import StrEnum  # import whichever version is used there
from undate.converters.base import BaseCalendarConverter, BaseDateConverter
from undate.date import Date, DatePrecision, Timedelta, UnDelta, UnInt


class TestUndate:
    def test_str(self):
        assert str(Undate(2022, 11, 7)) == "2022-11-07"
        assert str(Undate(2022, 11)) == "2022-11"
        assert str(Undate(2022)) == "2022"
        assert str(Undate(month=11, day=7)) == "--11-07"

    def test_partially_known_str(self):
        assert str(Undate("19XX")) == "19XX"
        assert str(Undate(2022, "1X")) == "2022-1X"
        assert str(Undate(2022, 11, "2X")) == "2022-11-2X"
        assert str(Undate(month="1X", day=7)) == "--1X-07"

        # TODO: should not allow initializing year/day without month;
        # should we infer unknown month? or raise an exception?
        # assert str(Undate(2022, day="2X")) == "2022-XX-2X"  # currently returns 2022-2X
        # assert str(Undate(2022, day=7)) == "2022-XX-07"   @ currently returns 2022-07

    def test_repr(self):
        # import undate to test eval of fully-qualified undate repr string
        import undate  # noqa: F401

        nov2022 = Undate(2022, 11, 7)
        # repr string should provide sufficient details to initialize
        assert (
            repr(nov2022)
            == "undate.Undate(year=2022, month=11, day=7, calendar='Gregorian')"
        )
        # eval on repr string should be equivalent to the object
        assert eval(repr(nov2022)) == nov2022
        nov2022_labeled = Undate(2022, 11, 7, label="A Special Day")
        assert (
            repr(nov2022_labeled)
            == "undate.Undate(year=2022, month=11, day=7, label='A Special Day', calendar='Gregorian')"
        )
        assert eval(repr(nov2022_labeled)) == nov2022_labeled
        # different calendar, missing fields
        islamic_date = Undate(484, calendar=Calendar.ISLAMIC)
        assert repr(islamic_date) == "undate.Undate(year=484, calendar='Islamic')"
        assert eval(repr(islamic_date)) == islamic_date

        # test string values for month/day
        unknown_year = Undate(month="1X", day="3X")
        assert (
            repr(unknown_year)
            == "undate.Undate(month='1X', day='3X', calendar='Gregorian')"
        )
        # unknown dates aren't equal, but string representation should match
        assert str(eval(repr(unknown_year))) == str(unknown_year)

    def test_init_str(self):
        assert Undate("2000").earliest.year == 2000
        # single or double digit string month should be ok
        assert Undate("2000", "2").earliest.month == 2
        assert Undate("2000", "02").earliest.month == 2

    def test_init_partially_known_year(self):
        uncertain1900s = Undate("19XX")
        assert uncertain1900s.earliest.year == 1900
        assert uncertain1900s.latest.year == 1999

        uncertain1x = Undate("1X05")
        assert uncertain1x.earliest.year == 1005
        assert uncertain1x.latest.year == 1905

        uncertain18x7 = Undate("18X7")
        assert uncertain18x7.earliest.year == 1807
        assert uncertain18x7.latest.year == 1897

    def test_init_partially_known_month(self):
        uncertain_fall = Undate(1900, "1X")
        assert uncertain_fall.earliest.month == 10
        assert uncertain_fall.latest.month == 12

        uncertain_notfall = Undate(1900, "0X")
        assert uncertain_notfall.earliest.month == 1
        assert uncertain_notfall.latest.month == 9

        # unlikely case, but now possible to calculate
        assert Undate(1900, "X1").earliest.month == 1
        assert Undate(1900, "X1").latest.month == 11

        # treat as unknown but allow
        unknown_month = Undate(1900, "XX")
        assert unknown_month.earliest.month == 1
        assert unknown_month.latest.month == 12
        assert str(unknown_month) == "1900-XX"

    def test_init_partially_known_day(self):
        uncertain_day = Undate(1900, 1, "XX")  # treat as None
        assert uncertain_day.earliest.day == 1
        assert uncertain_day.latest.day == 31

        uncertain_day = Undate(1900, 1, "1X")
        assert uncertain_day.earliest.day == 10
        assert uncertain_day.latest.day == 19

        uncertain_day = Undate(1900, 1, "0X")
        assert uncertain_day.earliest.day == 1
        assert uncertain_day.latest.day == 9
        uncertain_day = Undate(1900, 1, "2X")
        assert uncertain_day.earliest.day == 20
        assert uncertain_day.latest.day == 29
        uncertain_day = Undate(1900, 1, "3X")
        assert uncertain_day.earliest.day == 30
        assert uncertain_day.latest.day == 31

        uncertain_day = Undate(1900, 1, "X5")
        assert uncertain_day.earliest.day == 5
        assert uncertain_day.latest.day == 25

        uncertain_day = Undate(1900, 1, "X1")
        assert uncertain_day.earliest.day == 1
        assert uncertain_day.latest.day == 31

        # month with only 30 days
        uncertain_day = Undate(1900, 6, "X1")
        assert uncertain_day.earliest.day == 1
        assert uncertain_day.latest.day == 21  # doesn't go to 31
        uncertain_day = Undate(1900, 6, "3X")
        assert uncertain_day.earliest.day == 30
        assert uncertain_day.latest.day == 30

        # special cases
        # february! 28 days usually
        uncertain_day = Undate(1900, 2, "2X")
        assert uncertain_day.earliest.day == 20
        assert uncertain_day.latest.day == 28
        # february in a leap year
        uncertain_day = Undate(2024, 2, "2X")
        assert uncertain_day.latest.day == 29

        # TODO: handle leap day in an unknown year
        # (currently causes an exception because min/max years are not leap years)
        # Undate(None, 2, 29)

    def test_calendar(self):
        assert Undate(2024).calendar == Calendar.GREGORIAN
        # by name, any case
        assert Undate(848, calendar="ISLAMIC").calendar == Calendar.ISLAMIC
        assert Undate(848, calendar="islamic").calendar == Calendar.ISLAMIC
        # by enum
        assert Undate(848, calendar=Calendar.ISLAMIC).calendar == Calendar.ISLAMIC
        # invalid
        with pytest.raises(ValueError, match="Calendar `foobar` is not supported"):
            Undate(848, calendar="foobar")

    def test_as_calendar(self):
        # changes calendar *without* converting dates
        assert Undate(1243, 5, 7).as_calendar(Calendar.ISLAMIC) == Undate(
            1243, 5, 7, calendar=Calendar.ISLAMIC
        )
        # should also work with string
        assert Undate(1243, 5, 7).as_calendar("islamic") == Undate(
            1243, 5, 7, calendar=Calendar.ISLAMIC
        )

    def test_init_invalid(self):
        with pytest.raises(ValueError):
            Undate("19??")

        with pytest.raises(ValueError, match="At least one of year, month, or day"):
            Undate()

    def test_invalid_date(self):
        # invalid month should raise an error
        with pytest.raises(ValueError):
            Undate(1990, 22)

    def test_to_undate(self):
        undate_from_date = Undate.to_undate(date(2001, 3, 5))
        assert isinstance(undate_from_date, Undate)
        assert undate_from_date == Undate(2001, 3, 5)

        now = datetime.now()
        undate_from_dt = Undate.to_undate(now)
        assert isinstance(undate_from_dt, Undate)
        assert undate_from_dt == Undate(now.year, now.month, now.day)

        # from internal Date object
        y2k = Date(2000)
        y2k_to_undate = Undate.to_undate(y2k)
        assert isinstance(y2k_to_undate, Undate)
        assert int(y2k_to_undate.year) == y2k.year
        assert y2k_to_undate.month is None
        assert y2k_to_undate.day is None

        # unsupported type
        with pytest.raises(TypeError):
            Undate.to_undate("foo")

    # test properties for accessing parts of date
    def test_year_property(self):
        # two, three, four five digit years; numeric and string
        assert Undate(33).year == "0033"
        assert Undate(567).year == "0567"
        assert Undate("1984").year == "1984"
        assert Undate(23055).year == "23055"
        # partially known year
        assert Undate("19XX").year == "19XX"
        # unset year
        assert Undate(month=12, day=31).year == "XXXX"

        # NOTE: no longer supported to initialize undate with no date information
        # force method to hit conditional for date precision
        some_century = Undate(year="X")
        some_century.initial_values["year"] = None
        some_century.precision = DatePrecision.CENTURY
        assert some_century.year is None

    def test_month_property(self):
        # one, two digit month
        assert Undate(2023, 1).month == "01"
        assert Undate(2023, 12).month == "12"
        # partially unknown month (first/second digit unknown)
        assert Undate(2023, "1X").month == "1X"
        assert Undate(2023, "X2").month == "X2"
        # fully unknown month
        assert Undate(2023, "XX").month == "XX"
        # unset month, year precision
        assert Undate(2023).month is None
        # unset month, day precision (= some unknown month, not no month)
        assert Undate(day=15).month == "XX"

    def test_day_property(self):
        # one, two digit day
        assert Undate(2023, 1, 9).day == "09"
        assert Undate(2023, 1, 31).day == "31"
        # partially unknown day
        assert Undate(2023, 1, "1X").day == "1X"
        assert Undate(2023, 1, "X5").day == "X5"
        # fully unknown day
        assert Undate(2023, 1, "XX").day == "XX"
        # unset day
        assert Undate(2023).day is None
        assert Undate(2023, 1).day is None
        # Day without year or month
        assert Undate(day=15).day == "15"

        # force str based on date precision without day part set
        someday = Undate(2023)
        someday.precision = DatePrecision.DAY
        assert someday.day == "XX"

    def test_eq(self):
        assert Undate(2022) == Undate(2022)
        assert Undate(2022, 10) == Undate(2022, 10)
        assert Undate(2022, 10, 1) == Undate(2022, 10, 1)
        # dates without a known year cannot known to be equal
        assert not Undate(month=2, day=7) == Undate(month=2, day=7)

        # something we can't convert for comparison should return NotImplemented
        assert Undate(2022).__eq__("not a date") == NotImplemented

    def test_eq_datetime_date(self):
        # support comparisons with datetime objects for full day-precision
        assert Undate(2022, 10, 1) == date(2022, 10, 1)
        assert Undate(2022, 10, 1) != date(2022, 10, 2)
        assert Undate(1980, 10, 1) != date(2022, 10, 1)

        # other date precisions are not equal
        assert Undate(2022) != date(2022, 10, 1)
        assert Undate(2022, 10) != date(2022, 10, 1)

    def test_not_eq(self):
        assert Undate(2022) != Undate(2023)
        assert Undate(2022, 10) != Undate(2022, 11)
        assert Undate(2022, 10, 1) != Undate(2022, 10, 2)
        assert Undate(month=2, day=7) != Undate(month=2, day=8)
        assert Undate(2022) != Undate(2022, 10)
        assert Undate(2022, 10) != Undate(2022, 10, 1)

        # partially unknown dates should NOT be considered equal
        assert Undate("19XX") != Undate("19XX")
        assert Undate(1980, "XX") != Undate(1980, "XX")
        # same dates with unknown years should not be considered equal
        assert Undate(month=2, day=7) != Undate(month=2, day=7)

    testdata_lt_gt = [
        # dates to test for gt/lt comparison: earlier date, later date
        # - simple cases: same precision where one date is clearly earlier
        (Undate(2022), Undate(2023)),
        (Undate(1991, 1), Undate(1991, 5)),
        (Undate(1856, 3, 3), Undate(1856, 3, 21)),
        # - mixed precision where one date is clearly earlier
        (Undate(1991, 1), Undate(2000)),
        (Undate(1856, 3, 3), Undate(1901)),
        # partially known digits where comparison is possible
        (Undate("19XX"), Undate("20XX")),
        (Undate(1900, "0X"), Undate(1900, "1X")),
        # compare with datetime.date objects
        (Undate("19XX"), date(2020, 1, 1)),
        (Undate(1991, 1), date(1992, 3, 4)),
    ]

    @pytest.mark.parametrize("earlier,later", testdata_lt_gt)
    def test_lt(self, earlier, later):
        assert earlier < later
        assert later > earlier

    testdata_lte_gte = testdata_lt_gt.copy()
    # add a few exactly equal cases
    testdata_lte_gte.extend(
        [
            (Undate(1601), Undate(1601)),
            (Undate(1991, 1), Undate(1991, 1)),
            (Undate(1492, 5, 3), Undate(1492, 5, 3)),
            # compare with datetime.date also
            (Undate(1492, 5, 3), date(1492, 5, 3)),
        ]
    )

    def test_lt_when_eq(self):
        # strict less than / greater should return false when equal
        assert not Undate(1900) > Undate(1900)
        assert not Undate(1900) < Undate(1900)
        # same for datetime.date
        assert not Undate(1903, 1, 5) < date(1903, 1, 5)
        assert not Undate(1903, 1, 5) > date(1903, 1, 5)

    @pytest.mark.parametrize("earlier,later", testdata_lte_gte)
    def test_lte(self, earlier, later):
        assert earlier <= later
        assert later >= earlier

    def test_gt_lt_unknown_years(self):
        # unknown years cannot be compared on either side...
        year100 = Undate(100)
        some_january = Undate(month=1)
        assert not year100 < some_january
        assert not year100 <= some_january
        assert not year100 > some_january
        assert not year100 >= some_january
        assert not some_january < year100
        assert not some_january <= year100
        assert not some_january > year100
        assert not some_january >= year100

    def test_lt_notimplemented(self):
        # unsupported type should bail out and return NotImplemented
        assert Undate(2022).__lt__("foo") == NotImplemented

        # how to compare mixed precision where dates overlap?
        # if the second date falls *within* earliest/latest,
        # then it is not clearly less; not implemented?
        with pytest.raises(
            NotImplementedError,
            match="one date \\(2022-05\\) falls within the other \\(2022\\)",
        ):
            assert Undate(2022) < Undate(2022, 5)

        # same if we attempt to compare in the other direction
        with pytest.raises(
            NotImplementedError,
            match="one date \\(2022-05\\) falls within the other \\(2022\\)",
        ):
            assert Undate(2022, 5) < Undate(2022)

    testdata_contains = [
        # first date falls within the range of the other
        # dates within range: middle, start, end, varying precision
        (Undate(2022, 6), Undate(2022)),
        (Undate(2022, 1, 1), Undate(2022)),
        (Undate(2022, 12, 31), Undate(2022)),
        (Undate(2022, 6, 15), Undate(2022, 6)),
        # support contains with datetime.date
        (date(2022, 6, 1), Undate(2022)),
        (date(2022, 6, 1), Undate(2022, 6)),
    ]

    @pytest.mark.parametrize("date1,date2", testdata_contains)
    def test_contains(self, date1, date2):
        assert date1 in date2

        # unsupported type should bail out and return NotImplemented
        assert Undate(2022).__contains__("foo") == NotImplemented

    testdata_not_contains = [
        # dates not in range
        (Undate(1980), Undate(2020)),
        (Undate(1980), Undate(2020, 6)),
        (Undate(1980, 6), Undate(2020, 6)),
        # support contains with datetime.date
        (date(1980, 6, 1), Undate(2022)),
        (date(3001, 6, 1), Undate(2022, 6)),
        # partially known dates that are similar but same precision,
        # so one does not contain the other
        (Undate("199X"), Undate("19XX")),
        # - specific month to unknown month
        (Undate(1980, 6), Undate(1980, "XX")),
        # some of these might overlap, but we don't have enough
        # information to determine
        # - unknown month to unknown month
        (Undate(1980, "XX"), Undate(1980, "XX")),
        # - partially unknown month to unknown month
        (Undate(1801, "1X"), Undate(1801, "XX")),
        # fully unknown year
        (Undate(month=6, day=1), Undate(2022)),
        (Undate(1950), Undate(day=31)),
    ]

    @pytest.mark.parametrize("date1,date2", testdata_not_contains)
    def test_not_contains(self, date1, date2):
        assert date1 not in date2

    def test_sorting(self):
        # sorting should be possible based on gt/lt
        # test simple cases for sorting
        d1980 = Undate(1980)
        d2002_10 = Undate(2002, 10)
        d2002_12 = Undate(2002, 12)
        d2012_05_01 = Undate(2012, 5, 1)

        assert sorted([d2012_05_01, d2002_12, d2002_10, d1980]) == [
            d1980,
            d2002_10,
            d2002_12,
            d2012_05_01,
        ]

        # what about semi-ambigous cases?
        d1991_XX = Undate(1991, "XX")
        d1992_01_XX = Undate(1992, 1, "XX")
        assert sorted([d1992_01_XX, d1991_XX, d1980]) == [d1980, d1991_XX, d1992_01_XX]

        # what about things we can't compare?
        d1991 = Undate(1991)
        d1991_02 = Undate(1991, 2)
        # for now, this will raise a not implemented error
        with pytest.raises(NotImplementedError):
            sorted([d1991_02, d1991, d1991_XX])

        # TODO: partially known year?
        # someyear = Undate("1XXX")
        # assert sorted([d1991, someyear]) == [someyear, d1991]

    def test_possible_years(self):
        assert Undate(1991).possible_years == [1991]
        assert Undate("190X").possible_years == range(1900, 1910)
        assert Undate("19XX").possible_years == range(1900, 2000)
        # uses step when missing digit is not last digit
        assert Undate("19X1").possible_years == range(1901, 1992, 10)
        assert Undate("2X25").possible_years == range(2025, 2926, 100)
        assert Undate("1XXX").possible_years == range(1000, 2000)
        # completely unknown year raises value error, because the range is not useful
        with pytest.raises(
            ValueError, match="cannot be returned for completely unknown year"
        ):
            assert Undate("XXXX").possible_years

        # non-gregorian years should return in original calendar
        assert Undate(401, calendar="Hebrew").possible_years == [401]

    def test_representative_years(self):
        # single year is returned as is
        assert Undate("1991").representative_years == [1991]
        # for an uncertain year, returns first leap year and non-leap year in range
        assert Undate("190X").representative_years == [1900, 1904]
        assert Undate("19XX").representative_years == [1900, 1904]
        # works for other calendars
        assert Undate("481X", calendar="Hebrew").representative_years == [
            4810,
            4811,
            4812,
            4813,
            4816,
            4818,
        ]

        # use mock to simulate a calendar without representative years filtering
        with mock.patch(
            "undate.converters.calendars.HebrewDateConverter.representative_years"
        ) as mock_representative_years:
            mock_representative_years.side_effect = NotImplementedError
            assert Undate("481X", calendar="Hebrew").representative_years == list(
                range(4810, 4820)
            )

    def test_duration(self):
        day_duration = Undate(2022, 11, 7).duration()
        assert isinstance(day_duration, Timedelta)
        assert day_duration.days == 1

        january_duration = Undate(2022, 1).duration()
        assert january_duration.days == 31
        feb_duration = Undate(2022, 2).duration()
        assert feb_duration.days == 28
        # 2024 is a known leap year
        leapyear_feb_duration = Undate(2024, 2).duration()
        assert leapyear_feb_duration.days == 29

        year_duration = Undate(2022).duration()
        assert year_duration.days == 365
        leapyear_duration = Undate(2024).duration()
        assert leapyear_duration.days == 366

    def test_duration_month_nongregorian(self):
        # known-months for non-gregorian calendars should not be uncertain
        assert Undate(1288, 4, calendar="Seleucid").duration().days == 29
        assert Undate(1548, 5, calendar="Seleucid").duration().days == 30
        assert Undate(4791, 11, calendar="Hebrew").duration().days == 30
        assert Undate(4808, 10, calendar="Hebrew").duration().days == 29
        assert Undate(942, 1, calendar="Islamic").duration().days == 30
        assert Undate(984, 8, calendar="Islamic").duration().days == 29

        # in some cases month length may vary by year
        assert Undate(month=4, calendar="Seleucid").duration().days == 29
        assert Undate(month=8, calendar="Hebrew").duration().days == UnInt(29, 30)
        assert Undate(month=1, calendar="Islamic").duration().days == 30

    def test_partiallyknown_duration(self):
        # day in unknown month/year
        # assert Undate(day=5).duration().days == 1
        assert Undate(day=5).duration().days == 1
        assert Undate(year=1900, month=11, day="2X").duration().days == 1

        # month in unknown year
        assert Undate(month=6).duration().days == 30
        # partially known month
        # 1X = October, November, or December = 30 or 31 days
        # should return a Undelta object
        unknown_month_duration = Undate(year=1900, month="1X").duration()
        assert isinstance(unknown_month_duration, UnDelta)
        assert unknown_month_duration.days == UnInt(30, 31)

        # completely unknown month should also return a Undelta object
        unknown_month_duration = Undate(year=1900, month="XX").duration()
        assert isinstance(unknown_month_duration, UnDelta)
        # possible range is 28 to 31 days
        assert unknown_month_duration.days == UnInt(28, 31)

        # the number of days in February of an unknown year is uncertain, since
        # it could vary with leap years; either 28 or 29 days
        feb_duration = Undate(month=2).duration()
        assert isinstance(feb_duration, UnDelta)
        assert feb_duration.days == UnInt(28, 29)

    def test_partiallyknownyear_duration(self):
        assert Undate("190X").duration().days == UnInt(365, 366)
        assert Undate("XXXX").duration().days == UnInt(365, 366)
        # if possible years don't include any leap years, duration is not ambiguous
        assert Undate("19X1").duration().days == 365
        # year duration logic should work in other calendars
        # islamic
        assert Undate("108X", calendar="Islamic").duration().days == UnInt(354, 355)
        # completely unknown years is calculated based on representative years
        assert Undate("XXXX", calendar="Islamic").duration().days == UnInt(354, 355)
        assert Undate("536X", calendar="Hebrew").duration().days == UnInt(353, 385)
        # different set of years could vary
        assert Undate("53X2", calendar="Hebrew").duration().days == UnInt(354, 385)
        # fully unknown year also works for Hebrew calendar
        assert Undate("XXX", calendar="Hebrew").duration().days == UnInt(353, 385)

    def test_known_year(self):
        # known OR partially known
        assert Undate(2022).known_year is True
        assert Undate(month=2, day=5).known_year is False
        # partially known year is not known
        assert Undate("19XX").known_year is False
        # fully known string year should be known
        assert Undate("1900").known_year is True

    def test_is_known_month(self):
        assert Undate(2022).is_known("month") is False
        assert Undate(2022, 2).is_known("month") is True
        assert Undate(2022, "5").is_known("month") is True
        assert Undate(2022, "1X").is_known("month") is False
        assert Undate(2022, "XX").is_known("month") is False

    def test_is_known_day(self):
        assert Undate(1984).is_known("day") is False
        assert Undate(month=1, day=3).is_known("day") is True
        assert Undate(month=1, day="5").is_known("day") is True
        assert Undate(month=1, day="X5").is_known("day") is False
        assert Undate(month=1, day="XX").is_known("day") is False

    def test_unknown_year(self):
        # fully unknown year
        assert Undate(month=2, day=5).unknown_year is True
        # known or partially known years = all false for unknown
        assert Undate(2022).unknown_year is False
        # partially known year is not unknown
        assert Undate("19XX").unknown_year is False
        # fully known string year should be known
        assert Undate("1900").unknown_year is False

    def test_is_unknown_month(self):
        # fully unknown month
        assert Undate(2022).is_unknown("month") is True
        assert Undate(day=10).is_unknown("month") is True
        assert Undate(2022, 2).is_unknown("month") is False
        assert Undate(2022, "5").is_unknown("month") is False
        assert Undate(2022, "1X").is_unknown("month") is False
        assert Undate(2022, "XX").is_unknown("month") is False

    def test_is_unknown_day(self):
        # fully unknown day
        assert Undate(1984).is_unknown("day") is True
        assert Undate(month=5).is_unknown("day") is True
        assert Undate(month=1, day=3).is_unknown("day") is False
        assert Undate(month=1, day="5").is_unknown("day") is False
        assert Undate(month=1, day="X5").is_unknown("day") is False
        assert Undate(month=1, day="XX").is_unknown("day") is False

    def test_parse(self):
        assert Undate.parse("1984", "EDTF") == Undate(1984)
        assert Undate.parse("1984-04", "EDTF") == Undate(1984, 4)
        assert Undate.parse("1984-04", "EDTF") == Undate(1984, 4)
        assert Undate.parse("2000/2001", "EDTF") == UndateInterval(
            Undate(2000), Undate(2001)
        )

        assert Undate.parse("1984", "ISO8601") == Undate(1984)
        assert Undate.parse("1984-04", "ISO8601") == Undate(1984, 4)
        # dates with unknown year are not equal; compare repr string
        assert repr(Undate.parse("--12-31", "ISO8601")) == repr(
            Undate(month=12, day=31)
        )

        # unsupported format
        with pytest.raises(ValueError, match="Unsupported format"):
            Undate.parse("1984", "foobar")
        with pytest.raises(ValueError, match="Unsupported format"):
            Undate.parse("1984", "%Y-%m")

    def test_format(self):
        # EDTF format
        assert Undate(1984).format("EDTF") == "1984"
        assert Undate(1984, 4).format("EDTF") == "1984-04"
        assert Undate(1984, 4, 15).format("EDTF") == "1984-04-15"
        assert Undate("19XX").format("EDTF") == "19XX"
        assert Undate(1984, "XX").format("EDTF") == "1984-XX"
        assert Undate(1984, 4, "XX").format("EDTF") == "1984-04-XX"
        assert Undate(month=12, day=31).format("EDTF") == "XXXX-12-31"

        # ISO8601 format
        assert Undate(1984).format("ISO8601") == "1984"
        assert Undate(1984, 4).format("ISO8601") == "1984-04"
        assert Undate(1984, 4, 15).format("ISO8601") == "1984-04-15"
        assert Undate(month=12, day=31).format("ISO8601") == "--12-31"

        # unsupported format
        with pytest.raises(ValueError, match="Unsupported format"):
            Undate(1984).format("foobar")
        with pytest.raises(ValueError, match="Unsupported format"):
            Undate(1984).format("%Y-%m")


def test_calendar_get_converter():
    # ensure we can retrieve a calendar converter for each
    # calendar named in our calendar enum
    for cal in Calendar:
        converter = Calendar.get_converter(cal)
        assert isinstance(converter, BaseCalendarConverter)
        assert converter.name.lower() == cal.name.lower()

    class BogusCalendar(StrEnum):
        """Unsupported calendars"""

        FOOBAR = auto()
        DUMMY = auto()

    # test error handling
    # ensure we raise a ValueError when an invalid calendar is requested
    with pytest.raises(ValueError, match="Unknown calendar"):
        Calendar.get_converter(BogusCalendar.FOOBAR)

    class DummyFormatter(BaseDateConverter):
        name = "Dummy"

    # also error if you request a converter that is not a calendar converter
    # NOTE: this fails because get_converter converts the enum to title case...
    # can't be tested with any of the existing non-calendar converters
    with pytest.raises(
        ValueError, match="Requested converter 'Dummy' is not a CalendarConverter"
    ):
        Calendar.get_converter(BogusCalendar.DUMMY)



import pytest
from lark.exceptions import UnexpectedCharacters, UnexpectedEOF

from undate.converters.calendars.hebrew.parser import hebrew_parser


# for now, just test that valid dates can be parsed

testcases = [
    # year
    "5362",
    # month + year
    # - with and without accent
    "Ṭevet 5362",
    "Tevet 5362",
    "Elul 4932",
    "Sivan 5581",
    # variant month name, with or without accent
    "Ṭeveth 5362",
    "Teveth 5362",
    "Iyyar 1526",
    "Iyar 1526",
    # day month year
    "26 Tammuz 4816",
    "7 Heshvan 5425",
    "26 Tishrei 5416",
    "26 Tishri 5416",
    "14 Adar 5403",
    "14 Adar I 5403",
    "9 Adar II 5404",
    "9 Adar Bet 5404",
    # two and 1 digit years
    "536",
    "53",
    "3",
]


@pytest.mark.parametrize("date_string", testcases)
def test_should_parse(date_string):
    assert hebrew_parser.parse(date_string)


error_cases = [
    # invalid days
    ("0 Tammuz 5403", UnexpectedCharacters),
    ("31 Tishri 5403", UnexpectedCharacters),
    # month alone
    ("Tishri", UnexpectedEOF),
    # month day only
    ("12 Heshvan", UnexpectedEOF),
    # invalid month
    ("Foo 383", UnexpectedCharacters),
    # wrong format
    ("2024-10-02", UnexpectedCharacters),
    # year month day not supported
    ("5403 Adar", UnexpectedCharacters),
    ("5403 Adar 14", UnexpectedCharacters),
]


@pytest.mark.parametrize("date_string,exception", error_cases)
def test_should_error(date_string, exception):
    with pytest.raises(exception):
        hebrew_parser.parse(date_string)



from lark import Token, Transformer, Tree

from undate import Undate, UndateInterval


class EDTFTransformer(Transformer):
    """transform edtf parse tree and return Undate or UndateInterval"""

    INT = int

    def timeinterval(self, items):
        # transformed result from parser should be two undate objects;
        # combine into an interval
        return UndateInterval(*items)

    def date(self, items):
        parts = {}
        for child in items:
            if child.data in ["year", "month", "day"]:
                # in each case we expect one value;
                # anonymous tokens convert to their value
                value = child.children[0]
                # convert to integer when possible; otherwise pass as string
                try:
                    value = int(value)
                except ValueError:
                    value = str(value)
                parts[str(child.data)] = value

        return Undate(**parts)

    def extended_interval(self, items):
        # same as level 1 time interval, except one item may be None
        # for an open-ended range
        return self.timeinterval(items)

    def unknown_date(self, token):
        # unknown date for interval should be passed in as None
        return None

    def get_values(self, items):
        # get a list of values from tokens; recurses to get subtree tokens
        values = []
        for i in items:
            if isinstance(i, Token):
                values.append(str(i))
            if isinstance(i, Tree):
                values.extend(self.get_values(i.children))
        return values

    def year_unspecified(self, items):
        # combine parts (numeric & unknown) into a single string
        value = "".join(self.get_values(items))
        return Tree(data="year", children=[value])

    def month_unspecified(self, items):
        # combine multiple parts into a single string
        value = "".join(self.get_values(items))
        return Tree(data="month", children=[value])

    def day_unspecified(self, items):
        # combine multiple parts into a single string
        value = "".join(self.get_values(items))
        return Tree(data="day", children=[value])

    def date_level1(self, items):
        return self.date(items)

    # year (including negative years) use default transformation

    def year_fivedigitsplus(self, items):
        # strip off the leading Y and convert to integer
        token = items[0]
        year = int(token.value.lstrip("Y"))
        return Tree(data="year", children=[year])



import calendar
import datetime

import pytest

from undate import Undate, UndateInterval
from undate.date import Timedelta


class TestUndateInterval:
    def test_init_types(self):
        # datetime.date - autoconvert
        interval = UndateInterval(datetime.date(2022, 1, 1), None)
        assert isinstance(interval.earliest, Undate)
        interval = UndateInterval(None, datetime.date(2022, 1, 1))
        assert isinstance(interval.latest, Undate)

        # unsupported type should raise exception
        with pytest.raises(
            ValueError, match="earliest date 2022 cannot be converted to Undate"
        ):
            UndateInterval(2022, None)

        with pytest.raises(
            ValueError, match="latest date 1982 cannot be converted to Undate"
        ):
            UndateInterval(None, "1982")

    def test_init_validation(self):
        with pytest.raises(ValueError, match="invalid interval"):
            UndateInterval(Undate(2020), Undate(1010))

    def test_str(self):
        # 2022 - 2023
        assert str(UndateInterval(Undate(2022), Undate(2023))) == "2022/2023"
        # 2022 - 2023-05
        assert str(UndateInterval(Undate(2022), Undate(2023, 5))) == "2022/2023-05"
        # 2022-11-01 to 2022-11-07
        assert (
            str(UndateInterval(Undate(2022, 11, 1), Undate(2023, 11, 7)))
            == "2022-11-01/2023-11-07"
        )

    def test_format(self):
        interval = UndateInterval(Undate(2000), Undate(2001))
        assert interval.format("EDTF") == "2000/2001"
        assert interval.format("ISO8601") == "2000/2001"

        # Open-ended intervals
        open_start = UndateInterval(latest=Undate(2000))
        assert open_start.format("EDTF") == "../2000"
        assert open_start.format("ISO8601") == "/2000"

        open_end = UndateInterval(earliest=Undate(2000))
        assert open_end.format("EDTF") == "2000/.."
        assert open_end.format("ISO8601") == "2000/"

    def test_repr(self):
        # import undate to test eval of fully-qualified repr string
        import undate  # noqa: F401

        # interval with start and end
        closed_interval = UndateInterval(Undate(2022), Undate(2023))
        assert (
            repr(closed_interval)
            == f"undate.UndateInterval(earliest={repr(closed_interval.earliest)}, latest={repr(closed_interval.latest)})"
        )
        # should be able to evaluate repr string to get an equivalent object
        assert eval(repr(closed_interval)) == closed_interval
        # interval with a label
        fancy_epoch = UndateInterval(Undate(2022), Undate(2023), label="Fancy Epoch")
        assert (
            repr(fancy_epoch)
            == f"undate.UndateInterval(earliest={repr(fancy_epoch.earliest)}, latest={repr(fancy_epoch.latest)}, label='Fancy Epoch')"
        )
        assert eval(repr(fancy_epoch)) == fancy_epoch

        open_interval = UndateInterval(
            Undate(33),
        )
        assert (
            repr(open_interval)
            == f"undate.UndateInterval(earliest={repr(open_interval.earliest)})"
        )
        assert eval(repr(open_interval)) == open_interval

    def test_str_open_range(self):
        # 900 -
        assert str(UndateInterval(Undate(900))) == "0900/"
        # - 1900
        assert str(UndateInterval(latest=Undate(1900))) == "../1900"
        # - 1900-12
        assert str(UndateInterval(latest=Undate(1900, 12))) == "../1900-12"

    def test_eq(self):
        assert UndateInterval(Undate(2022), Undate(2023)) == UndateInterval(
            Undate(2022), Undate(2023)
        )
        assert UndateInterval(Undate(2022), Undate(2023, 5)) == UndateInterval(
            Undate(2022), Undate(2023, 5)
        )
        assert UndateInterval(Undate(2022, 5)) == UndateInterval(Undate(2022, 5))

    def test_eq_type_check(self):
        # doesn't currently support comparison with anything else
        interval = UndateInterval(Undate(900))
        # returns NotImplemented if comparison with this type is not supported
        assert interval.__eq__("foo") == NotImplemented

    def test_not_eq(self):
        assert UndateInterval(Undate(2022), Undate(2023)) != UndateInterval(
            Undate(2022), Undate(2024)
        )
        assert UndateInterval(Undate(2022), Undate(2023, 5)) != UndateInterval(
            Undate(2022), Undate(2023, 6)
        )
        assert UndateInterval(Undate(2022), Undate(2023, 5)) != UndateInterval(
            Undate(2022), Undate(2023)
        )
        assert UndateInterval(Undate(2022, 5)) != UndateInterval(Undate(2022, 6))

    def test_min_year_non_leapyear(self):
        assert not calendar.isleap(Undate.MIN_ALLOWABLE_YEAR)

    def test_duration(self):
        week_duration = UndateInterval(
            Undate(2022, 11, 1), Undate(2022, 11, 7)
        ).duration()
        assert isinstance(week_duration, Timedelta)
        assert week_duration.days == 7

        twomonths = UndateInterval(Undate(2022, 11), Undate(2022, 12)).duration()
        # november - december = 30 days + 31 days
        assert twomonths.days == 30 + 31

        twoyears = UndateInterval(Undate(2021), Undate(2022)).duration()
        assert twoyears.days == 365 * 2

        # special case: month/day with no year (assumes same year)
        week_noyear_duration = UndateInterval(
            Undate(None, 11, 1), Undate(None, 11, 7)
        ).duration()
        assert week_noyear_duration.days == 7
        # special case 2: month/day with no year, wrapping from december to january
        # (assumes sequential years)
        month_noyear_duration = UndateInterval(
            Undate(None, 12, 1), Undate(None, 1, 1)
        ).duration()
        assert month_noyear_duration.days == 32

        # real world test cases from Shakespeare and Company Project data;
        # second date is a year minus one day in the future
        month_noyear_duration = UndateInterval(
            Undate(None, 6, 7), Undate(None, 6, 6)
        ).duration()
        assert month_noyear_duration.days == 365

        # durations that span february in unknown years should assume
        # non-leap years
        jan_march_duration = UndateInterval(
            Undate(None, 2, 28), Undate(None, 3, 1)
        ).duration()
        assert jan_march_duration.days == 2

        # duration is not supported for open-ended intervals
        with pytest.raises(
            NotImplementedError, match="Cannot calculate.*open-ended interval"
        ):
            assert UndateInterval(Undate(2000), None).duration()

        # one year set and the other not currently raises not implemented error
        with pytest.raises(NotImplementedError):
            UndateInterval(Undate(2000), Undate(month=10)).duration()

    def test_intersection(self):
        century11th = UndateInterval(Undate(1001), Undate(1100))
        century20th = UndateInterval(Undate(1901), Undate(2000))
        # no intersection
        assert century11th.intersection(century20th) is None
        # should work in either direction
        assert century20th.intersection(century11th) is None

        decade1990s = UndateInterval(Undate(1990), Undate(1999))
        # intersection of an interval completely contained in another
        # returns an interval equivalent to the smaller one
        assert century20th.intersection(decade1990s) == decade1990s
        assert decade1990s.intersection(century20th) == decade1990s

        # partial overlap
        nineties_oughts = UndateInterval(Undate(1990), Undate(2009))
        assert century20th.intersection(nineties_oughts) == UndateInterval(
            Undate(1990), Undate(2000)
        )

        # intersections between half open intervals
        after_c11th = UndateInterval(Undate(1001), None)
        assert after_c11th.intersection(century20th) == century20th
        assert after_c11th.intersection(decade1990s) == decade1990s

        before_20th = UndateInterval(None, Undate(1901))
        assert before_20th.intersection(decade1990s) is None
        assert before_20th.intersection(century11th) == century11th
        assert before_20th.intersection(after_c11th) == UndateInterval(
            Undate(1001), Undate(1901)
        )

    def test_contains(self):
        century11th = UndateInterval(Undate(1001), Undate(1100))
        century20th = UndateInterval(Undate(1901), Undate(2000))
        decade1990s = UndateInterval(Undate(1990), Undate(1999))
        # an interval DOES contain itself
        for interval in [century11th, century20th, decade1990s]:
            assert interval in interval

        # checking if an interval is within another interval
        assert decade1990s in century20th
        assert decade1990s not in century11th
        assert century11th not in decade1990s
        assert century20th not in decade1990s
        # a specific date can be contained by an interval
        y2k = Undate(2000)
        assert y2k in century20th
        assert y2k not in century11th
        # partially known date should work too
        april_someyear = Undate("198X", 4)
        assert april_someyear in century20th
        assert april_someyear not in century11th
        # conversion from datetime.date also works
        assert datetime.date(1922, 5, 1) in century20th
        # unsupported types result in a type error
        with pytest.raises(TypeError):
            assert "nineteen-eighty-four" in century20th

        # contains check with half-open intervals
        after_c11th = UndateInterval(Undate(1001), None)
        before_20th = UndateInterval(None, Undate(1901))
        # neither of them contains the other
        assert after_c11th not in before_20th
        assert before_20th not in after_c11th
        # nor are they contained by a smaller range
        assert after_c11th not in decade1990s
        assert before_20th not in decade1990s

        # all of our previous test dates are in the 1900s,
        # so they are after the 11th century and not before the 20th
        for period in [decade1990s, y2k, april_someyear]:
            assert period in after_c11th
            assert period not in before_20th

        # fully open interval - is this even meaningful?
        whenever = UndateInterval(None, None)
        assert decade1990s in whenever
        # NOTE: an interval contains itself or an equivalent interval,
        # but that may not make sense for open intervals...
        assert whenever in whenever
